from __future__ import unicode_literals
from nose.plugins.attrib import attr
from base_tests import GraphPropertyBaseClassTestCase
from mogwai.properties.properties import IPV6, IPV6WithV4
from mogwai.models import Vertex
from mogwai._compat import print_


@attr('unit', 'property', 'property_ipv6')
class IPV6PropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = IPV6
    good_cases = ('1:2:3:4:5:6:7:8', '1::', '1:2:3:4:5:6:7::',
                  '1::8', '1:2:3:4:5:6::8', '1:2:3:4:5:6::8',
                  '1::7:8', '1:2:3:4:5::7:8', '1:2:3:4:5::8',
                  '1::6:7:8', '1:2:3:4::6:7:8', '1:2:3:4::8',
                  '1::5:6:7:8', '1:2:3::5:6:7:8', '1:2:3::8',
                  '1::4:5:6:7:8', '1:2::4:5:6:7:8', '1:2::8',
                  '1::3:4:5:6:7:8', '1::3:4:5:6:7:8', '1::8',
                  '::2:3:4:5:6:7:8', '::2:3:4:5:6:7:8', '::8', '::')
    bad_cases = ('0', '0.', '0.0', '0.0.', '0.0.0', '0.0.0.', '256.256.256.256', '1.2.3.256')

    def test_ipv6_default_cases(self):
        p = IPV6(default='1:2:3:4:5:6:7:8')


class IPV6TestVertex(Vertex):
    element_type = 'test_ipv6_vertex'

    test_val = IPV6()


@attr('unit', 'property', 'property_ipv6')
class IPV6VertexTestCase(GraphPropertyBaseClassTestCase):

    def test_ipv6_io(self):
        print_("creating vertex")
        dt = IPV6TestVertex.create(test_val='1::8')
        print_("getting vertex from vertex: %s" % dt)
        dt2 = IPV6TestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = IPV6TestVertex.create(test_val='1::7:8')
        print_("\ncreated vertex: %s" % dt)
        dt2 = IPV6TestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, '1::7:8')
        print_("deleting vertex")
        dt2.delete()


@attr('unit', 'property', 'property_ipv6w4')
class IPV6WithV4PropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = IPV6WithV4
    good_cases = ('1:2:3:4:5:6:7:8', '1::', '1:2:3:4:5:6:7::',  # IPv6
                  '1::8', '1:2:3:4:5:6::8', '1:2:3:4:5:6::8',
                  '1::7:8', '1:2:3:4:5::7:8', '1:2:3:4:5::8',
                  '1::6:7:8', '1:2:3:4::6:7:8', '1:2:3:4::8',
                  '1::5:6:7:8', '1:2:3::5:6:7:8', '1:2:3::8',
                  '1::4:5:6:7:8', '1:2::4:5:6:7:8', '1:2::8',
                  '1::3:4:5:6:7:8', '1::3:4:5:6:7:8', '1::8',
                  '::2:3:4:5:6:7:8', '::2:3:4:5:6:7:8', '::8', '::',
                  '::255.255.255.255', '::ffff:255.255.255.255', '::ffff:0:255.255.255.255',  # Mapped/Translated
                  '2001:db8:3:4::192.0.2.33', '64:ff9b::192.0.2.33',  # Embedded
                  )
    bad_cases = ('0', '0.', '0.0', '0.0.', '0.0.0', '0.0.0.', '256.256.256.256', '1.2.3.256', '1.2.3.4')

    def test_ipv6_default_cases(self):
        p = IPV6WithV4(default='1:2:3:4:5:6:7:8')


class IPV64TestVertex(Vertex):
    element_type = 'test_ipv64_vertex'

    test_val = IPV6WithV4()


@attr('unit', 'property', 'property_ipv6w4')
class IPV64VertexTestCase(GraphPropertyBaseClassTestCase):

    def test_ipv64_io(self):
        print_("creating vertex")
        dt = IPV64TestVertex.create(test_val='::255.255.255.255')
        print_("getting vertex from vertex: %s" % dt)
        dt2 = IPV64TestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = IPV64TestVertex.create(test_val='::ffff:255.255.255.255')
        print_("\ncreated vertex: %s" % dt)
        dt2 = IPV64TestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, '::ffff:255.255.255.255')
        print_("deleting vertex")
        dt2.delete()


from __future__ import unicode_literals
from mogwai.tests import BaseMogwaiTestCase
from mogwai.properties.strategy import *
from nose.plugins.attrib import attr


@attr('unit', 'strategy')
class StrategyBaseClassTestCase(BaseMogwaiTestCase):
    """ Test Base Strategy Callable Object """

    def setUp(self):
        self.klass = Strategy

    def test_callable(self):
        """ Test that the strategy is callable """
        self.assertTrue(callable(self.klass.condition))

    def test_strategy(self):
        """ Test Base Strategy Callable Object """
        self.assertRaises(NotImplementedError, self.klass.condition, 0, 1, True, False)


@attr('unit', 'strategy')
class SaveOnceStrategyTestCase(StrategyBaseClassTestCase):
    """ Test SaveOnce Strategy """

    def setUp(self):
        self.klass = SaveOnce

    def test_strategy(self):
        """ Test SaveOnce Strategy """
        self.assertTrue(self.klass.condition(previous_value=None,
                                             value=1,
                                             has_changed=False,
                                             first_save=True))
        self.assertRaises(SaveStrategyException, self.klass.condition, 0, 1, True, False)


@attr('unit', 'strategy')
class SaveAlwaysStrategyTestCase(StrategyBaseClassTestCase):
    """ Test SaveAlways Strategy """

    def setUp(self):
        self.klass = SaveAlways

    def test_strategy(self):
        """ Test SaveAlways Strategy """
        self.assertTrue(self.klass.condition(previous_value=None,
                                             value=1,
                                             has_changed=False,
                                             first_save=True))


@attr('unit', 'strategy')
class SaveOnChangeStrategyTestCase(StrategyBaseClassTestCase):
    """ Test SaveOnChange Strategy """

    def setUp(self):
        self.klass = SaveOnChange

    def test_strategy(self):
        """ Test SaveOnChange Strategy """
        self.assertTrue(self.klass.condition(previous_value=None,
                                             value=1,
                                             has_changed=True,
                                             first_save=True))
        self.assertFalse(self.klass.condition(previous_value=1,
                                              value=1,
                                              has_changed=False,
                                              first_save=False))


@attr('unit', 'strategy')
class SaveOnIncreaseStrategyTestCase(StrategyBaseClassTestCase):
    """ Test SaveIncrease Strategy """

    def setUp(self):
        self.klass = SaveOnIncrease

    def test_strategy(self):
        """ Test SaveIncrease Strategy """
        self.assertTrue(self.klass.condition(previous_value=None,
                                             value=1,
                                             has_changed=True,
                                             first_save=True))
        self.assertFalse(self.klass.condition(previous_value=1,
                                              value=1,
                                              has_changed=False,
                                              first_save=False))
        self.assertFalse(self.klass.condition(previous_value=2,
                                              value=1,
                                              has_changed=True,
                                              first_save=False))


@attr('unit', 'strategy')
class SaveOnDecreaseStrategyTestCase(StrategyBaseClassTestCase):
    """ Test SaveOnDecrease Strategy """

    def setUp(self):
        self.klass = SaveOnDecrease

    def test_strategy(self):
        """ Test SaveOnChange Strategy """
        self.assertFalse(self.klass.condition(previous_value=None,
                                              value=1,
                                              has_changed=True,
                                              first_save=True))
        self.assertFalse(self.klass.condition(previous_value=1,
                                              value=1,
                                              has_changed=False,
                                              first_save=False))
        self.assertTrue(self.klass.condition(previous_value=2,
                                             value=1,
                                             has_changed=True,
                                             first_save=False))


from __future__ import unicode_literals
from nose.plugins.attrib import attr
from mogwai.models import Vertex, Edge
from mogwai.properties import String, Integer
from mogwai.exceptions import MogwaiRelationshipException
from mogwai.tests.base import BaseMogwaiTestCase, TestVertexModel, counter
from mogwai.relationships.base import Relationship


class TestRelationshipEdgeModel(Edge):
    label = 'test_relationship_edge_model'

    name = String(default='test_edge')
    test_val = Integer(default=counter)


class TestRelationshipVertexModel(Vertex):
    element_type = 'test_relationship_vertex_model'

    name = String(default='test_vertex')
    test_val = Integer(default=counter)

    relation = Relationship(TestRelationshipEdgeModel, TestVertexModel)


@attr('unit', 'relationship')
class GraphRelationshipVertexIOTestCase(BaseMogwaiTestCase):
    """ Test Relationship Vertex IO Functionality """

    @classmethod
    def setUpClass(cls):
        super(GraphRelationshipVertexIOTestCase, cls).setUpClass()
        cls.relationship_base_cls = Relationship
        cls.edge_model = TestRelationshipEdgeModel
        cls.vertex_model = TestRelationshipVertexModel

    def test_instantiation(self):
        """ Test that the Relationship is properly Instantiated """

        v1 = self.vertex_model.create(name='test relationship')
        # setup relationship
        self.assertIsNotNone(v1.relation.top_level_vertex_class)
        self.assertIsNotNone(v1.relation.top_level_vertex)
        self.assertEqual(v1.relation.top_level_vertex, v1)
        v1.delete()

    def test_follow_through(self):
        """ Test that the Relationship property functions """

        v1 = self.vertex_model.create(name='test relationship')
        e1, v2 = v1.relation.create(vertex_params={'name': 'new relation'})

        e1q = v1.outE(TestRelationshipEdgeModel)[0]
        v2q = v1.outV(TestRelationshipEdgeModel)[0]

        self.assertEqual(e1, e1q)
        self.assertEqual(v2, v2q)

        e1.delete()
        v1.delete()
        v2.delete()


from __future__ import unicode_literals
from nose.plugins.attrib import attr
from mogwai.tests import BaseMogwaiTestCase
from base_tests import GraphPropertyBaseClassTestCase
from mogwai.properties.properties import String, Text, GraphProperty
from mogwai.models import Vertex
from mogwai.exceptions import ValidationError
from mogwai._compat import print_


@attr('unit', 'property', 'property_string')
class StringPropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = String
    good_cases = ('test', )
    bad_cases = (None, 0, 1.1, [], (), {})

    def test_max_min_length_validation_error(self):
        s = self.klass(max_length=5, min_length=2)
        with self.assertRaises(ValidationError):
            s.validate('123456')
        with self.assertRaises(ValidationError):
            s.validate('1')


class StringTestVertex(Vertex):
    element_type = 'test_string_vertex'

    test_val = String()

CHOICES = (
    ('A', 1),
    ('B', 2)
)


class StringTestChoicesVertex(Vertex):
    element_type = 'test_string_choices_vertex'

    test_val = String(choices=CHOICES)


@attr('unit', 'property', 'property_string')
class StringVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_string_io(self):
        print_("creating vertex")
        dt = StringTestVertex.create(test_val='test')
        print_("getting vertex from vertex: %s" % dt)
        dt2 = StringTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = StringTestVertex.create(test_val='tset')
        print_("\ncreated vertex: %s" % dt)
        dt2 = StringTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, 'tset')
        print_("deleting vertex")
        dt2.delete()


@attr('unit', 'property', 'property_string')
class TextPropertyTestCase(StringPropertyTestCase):
    klass = Text


class TextTestVertex(Vertex):
    element_type = 'test_text_vertex'

    test_val = Text()


@attr('unit', 'property', 'property_string')
class TextVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_text_io(self):
        print_("creating vertex")
        dt = TextTestVertex.create(test_val='ab12')
        print_("getting vertex from vertex: %s" % dt)
        dt2 = TextTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = TextTestVertex.create(test_val='12ab')
        print_("\ncreated vertex: %s" % dt)
        dt2 = TextTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, '12ab')
        print_("deleting vertex")
        dt2.delete()


@attr('unit', 'property', 'property_string')
class TestVertexChoicesTestCase(BaseMogwaiTestCase):

    def test_good_choices_value_io(self):
        print_("creating vertex")
        dt = StringTestChoicesVertex.create(test_val=1)
        print_("validating input")
        self.assertEqual(dt.test_val, 'A')
        print_("deleting vertex")
        dt.delete()

    def test_good_choices_key_io(self):
        print_("creating vertex")
        dt = StringTestChoicesVertex.create(test_val='B')
        print_("validating input")
        self.assertEqual(dt.test_val, 'B')
        print_("deleting vertex")
        dt.delete()

    def test_bad_choices_io(self):
        with self.assertRaises(ValidationError):
            print_("creating vertex")
            dt = StringTestChoicesVertex.create(test_val=3)
            print_("validating input")
            self.assertEqual(dt.test_val, 'C')
            print_("deleting vertex")
            dt.delete()


from __future__ import unicode_literals
from _compat import string_types, array_types
import logging
from re import compile
from random import shuffle
from rexpro.connection import RexProConnection
from rexpro.exceptions import RexProConnectionException, RexProScriptException

from mogwai.exceptions import MogwaiConnectionError, MogwaiQueryError
from mogwai.metrics.manager import MetricManager
from mogwai._compat import print_

logger = logging.getLogger(__name__)


_connections = []
_graph_name = None
metric_manager = MetricManager()
_loaded_models = []
__cached_spec = None


@metric_manager.time_calls
def execute_query(query, params={}, transaction=True, isolate=True, *args, **kwargs):
    """
    Execute a raw Gremlin query with the given parameters passed in.

    :param query: The Gremlin query to be executed
    :type query: str
    :param params: Parameters to the Gremlin query
    :type params: dict
    :param context: String context data to include with the query for stats logging
    :rtype: dict

    """
    if len(_connections) <= 0:  # pragma: no cover
        raise MogwaiConnectionError('Must call mogwai.connection.setup before querying.')

    conn = _connections[0]
    """ @type conn: rexpro.RexProConnection """
    try:
        response = conn.execute(query, params=params, isolate=isolate, transaction=transaction)
        print_("Got raw response: %s" % response)

    except RexProConnectionException as ce:  # pragma: no cover
        raise MogwaiConnectionError("Connection Error during query - {}".format(ce))
    except RexProScriptException as se:  # pragma: no cover
        raise MogwaiQueryError("Error during query - {}".format(se))
    except:  # pragma: no cover
        raise

    logger.debug(response)

    return response


_host_re = compile(r'^((?P<user>.+?)(:(?P<password>.*?))?@)?(?P<host>.*?)(:(?P<port>\d+?))?(?P<graph_name>/.*?)?$')


def _parse_host(host, username, password, graph_name):
        m = _host_re.match(host)
        d = m.groupdict() if m is not None else {}
        host = d.get('host', None) or '127.0.0.1'
        port = int(d.get('port', None) or 8184)
        username = d.get('username', None) or username
        password = d.get('password', None) or password
        graph_name = d.get('graph_name', None) or graph_name
        return {'host': host, 'port': port, 'username': username, 'password': password, 'graph_name': graph_name}


def setup(hosts, graph_name='graph', username='', password='', metric_reporters=None):
    """  Sets up the connection, and instantiates the models

    """
    global _connections
    global metric_manager

    if metric_reporters:  # pragma: no cover
        metric_manager.setup_reporters(metric_reporters)

    if isinstance(hosts, string_types):
        _connections.append(RexProConnection(**_parse_host(hosts, username, password, graph_name)))
    elif isinstance(hosts, array_types):  # pragma: no cover
        for host in hosts:
            _connections.append(RexProConnection(**_parse_host(host, username, password, graph_name)))

        shuffle(_connections)
    else:  # pragma: no cover
        raise MogwaiConnectionError("Must Specify at least one host or list of hosts")


def _add_model_to_space(model):
    global _loaded_models
    _loaded_models.append(model)


def generate_spec():
    """ Generates a titan index and type specification document based on loaded Vertex and Edge models """
    global _loaded_models, __cached_spec
    if __cached_spec:
        return __cached_spec
    from mogwai.models import Edge
    spec_list = []
    for model in _loaded_models:
        if not model.__abstract__ and hasattr(model, 'get_element_type'):

            makeType = 'makeLabel' if issubclass(model, Edge) else 'makeKey'
            element_type = 'Edge' if issubclass(model, Edge) else 'Vertex'

            spec = {'model': model.get_element_type(),
                    'element_type': element_type,
                    'makeType': makeType,
                    'properties': {}}
            for property in model._properties.values():
                if property.index:
                    # Only set this up for indexed properties

                    # Uniqueness constraint
                    uniqueness = ""
                    if property.unique and property.unique.lower() == 'in':
                        uniqueness = ".unique()"
                    elif property.unique and property.unique.lower() == 'out':
                        uniqueness = ".unidirected()"
                    elif property.unique and property.unique.lower() == 'both':
                        uniqueness = ".unique().single()"
                    if property.unique and property.unique.lower() == 'list':
                        uniqueness += ".list()"

                    # indexing extensions support
                    if not property.index_ext:
                        index_ext = ""
                    else:
                        index_ext = ".indexed(%s)" % property.index_ext

                    #compiled_index = {"script": "g.createKeyIndex(name, {}.class).dataType({}.class){}{}.make(); g.commit()".format(
                    #    element_type,
                    #    property.data_type,
                    #    index_ext,
                    #    uniqueness),
                    compiled_index = {"script": "g.{}(name).dataType({}.class).indexed({}{}.class){}.make(); g.commit()".format(
                        makeType,
                        property.data_type,
                        index_ext,
                        element_type,
                        uniqueness),
                                      "params": {'name': property.db_field_name},
                                      "transaction": False}
                    spec['properties'][property.db_field_name] = {
                        'data_type': property.data_type,
                        'index_ext': index_ext,
                        'uniqueness': uniqueness,
                        'compiled': compiled_index,
                    }

            spec_list.append(spec)
    __cached_spec = spec_list
    return spec_list


def sync_spec(filename, host, graph_name='graph', username='', password='', dry_run=False):  # pragma: no cover
    """
    Sync the given spec file to mogwai.

    @param filename: The filename of the spec file
        @type filename: str
    @param host: The host the be synced
        @type host: str
    @param graph_name: The name of the graph to be synced
        @type graph_name: str
    @param dry_run: Only prints generated Gremlin if True
        @type dry_run: boolean
    @returns: None

    """
    conn = RexProConnection(graph_name=graph_name, **_parse_host(host, username, password, graph_name))
    #Spec(filename).sync(conn, dry_run=dry_run)
    pass



from __future__ import unicode_literals
from pyformance.registry import MetricsRegistry, RegexRegistry
import time
from twisted.internet import task
from mogwai.exceptions import MogwaiMetricsException
from mogwai._compat import print_


def get_time():
    return int(round(time.time()))


class BaseMetricsReporter(object):
    """ Base Metrics Reporter class

    Customize this class for the specific metrics service, ie. Graphite, NewRelic, etc.
    """

    def __init__(self, registry=None, reportingInterval=5*60, metric_prefix=None):
        """ Create a Metrics Reporter

        To override this method, please make sure to call the superclass *after* your methods. Ie.
        def __init__(self, mystuff, stuff2=None, *args, **kwargs):
            # do my stuff...
            super(MyClass, self).__init__(*args, **kwargs)

        @param registry: The pyformance registry
            @type registry: list[ MetricsRegistry | RegexRegistry ] | MetricsRegistry | RegexRegistry
        @param reportingInterval: The interval (number of seconds) on which to report the collected metrics.
            @type reportingInterval: float | long | int
        @param metric_prefix: The prefix on the collected metrics. The default implementation ignores this.
            @type metric_prefix: basestring | None
        """
        self.setup_registry(registry=registry)
        self.reporting_interval = reportingInterval
        self.metric_prefix = metric_prefix or ''
        self.task = None

    def start(self):
        self.start_reporter(self.reporting_interval)

    def stop(self):
        self.task.stop()

    def setup_registry(self, registry=None):
        if not registry:
            self.registry = [MetricsRegistry(), ]
        else:
            if not isinstance(registry, (tuple, list)):
                registry = [registry, ]
            self.registry = []
            for reg in registry:
                if not isinstance(reg, (MetricsRegistry, RegexRegistry)):
                    raise MogwaiMetricsException(
                        "%s is not an instance of pyformance.MetricsRegistry or pyformance.RegexRegistry"
                    )
                self.registry.append(reg)

    def start_reporter(self, reportingInterval):
        """ Start the Metric Reporter Agent

        Kicks off a loop that every reportingInterval runs the `send_metrics` method.

        @param reportingInterval: The interval (number of seconds) on which to report the collected metrics.
            @type reportingInterval: float | long | int
        """
        self.task = task.LoopingCall(self.send_metrics)
        self.task.start(reportingInterval)

    def _get_metrics(self, timestamp=None, *args, **kwargs):
        """ Default pyformance implementation to collect all the metrics from the registries.

        Change only if you know what you are doing.

        @param timestamp: use this timestamp instead of the generated timestamp when the method is run
            @type timestamp: long | int
        @returns: Timestamp and aggregated metrics from all registries.
            @rtype: tuple( long | int, dict )
        """
        if not timestamp:
            timestamp = get_time()
        metrics = {}
        for reg in self.registry:
            metrics.update(reg.dump_metrics())
        return timestamp, metrics

    def get_metrics(self, timestamp=None, *args, **kwargs):
        """ Default pyformance implementation to collect all the metrics from the registries.

        Change this if you wan't customize for a particular metric collection system, ie. graphite, newrelic, etc.

        @param timestamp: use this timestamp instead of the generated timestamp when the method is run
            @type timestamp: long | int
        @returns: Timestamp and aggregated metrics from all registries.
            @rtype: tuple( long | int, dict )
        """
        return self._get_metrics(timestamp)

    def send_metrics(self, *args, **kwargs):
        """ Default pyformance implementation is to dump all metrics to STDOUT

        Change this if you want to customize for a particular metric collection system, ie. graphite, newrelic, etc.
        """
        timestamp, metrics = self.get_metrics()
        if metrics:
            print_("{}: {}".format(timestamp, metrics))


ConsoleMetricReporter = BaseMetricsReporter

__all__ = ['MetricsRegistry', 'RegexRegistry', 'Counter', 'Histogram', 'Meter', 'Timer', 'MogwaiMetricsException',
           'BaseMetricsReporter', 'ConsoleMetricReporter']



from __future__ import unicode_literals
import datetime
from pytz import utc
from decimal import Decimal as D
from nose.plugins.attrib import attr

from mogwai.properties import *
from mogwai.tests.base import BaseMogwaiTestCase


@attr('unit', 'value_manager')
class TestChangedProperty(BaseMogwaiTestCase):
    """
    Tests that the `changed` property works as intended
    """

    def test_string_update(self):
        """ Tests changes on string types """
        vm = String.value_manager(String(), 'str', strategy=SaveOnChange)
        self.assertFalse(vm.changed)
        vm.value = 'unicode'
        self.assertTrue(vm.changed)

    def test_string_inplace_update(self):
        """ Tests changes on string in place types """
        vm = String.value_manager(String(), 'str', strategy=SaveOnChange)
        self.assertFalse(vm.changed)
        vm.value += 's'
        self.assertTrue(vm.changed)

    def test_integer_update(self):
        """ Tests changes on integer types """
        vm = Integer.value_manager(Integer(), 5, strategy=SaveOnChange)
        self.assertFalse(vm.changed)
        vm.value = 4
        self.assertTrue(vm.changed)

    def test_integer_inplace_update(self):
        """ Tests changes on integer in place types """
        vm = Integer.value_manager(Integer(), 5, strategy=SaveOnChange)
        self.assertFalse(vm.changed)
        vm.value += 1
        self.assertTrue(vm.changed)

    def test_datetime_update(self):
        """ Tests changes on datetime types """
        now = datetime.datetime.now(tz=utc)
        vm = DateTime.value_manager(DateTime(), now, strategy=SaveOnChange)
        self.assertFalse(vm.changed)
        vm.value = now + datetime.timedelta(days=1)
        self.assertTrue(vm.changed)

    def test_decimal_update(self):
        """ Tests changes on decimal types """
        vm = Decimal.value_manager(Decimal(), D('5.00'), strategy=SaveOnChange)
        self.assertFalse(vm.changed)
        vm.value = D('4.00')
        self.assertTrue(vm.changed)

    def test_decimal_inplace_update(self):
        """ Tests changes on decimal in place types """
        vm = Decimal.value_manager(Decimal(), D('5.00'), strategy=SaveOnChange)
        self.assertFalse(vm.changed)
        vm.value += D('1.00')
        self.assertTrue(vm.changed)

    def test_dictionary_update(self):
        """ Tests changes on dict types """
        vm = Dictionary.value_manager(Dictionary(), {1: 2, 3: 4}, strategy=SaveOnChange)
        self.assertFalse(vm.changed)
        vm.value = {4: 5}
        self.assertTrue(vm.changed)

    def test_dictionary_inplace_update(self):
        """ Tests changes on dict in place types """
        vm = Dictionary.value_manager(Dictionary(), {1: 2, 3: 4}, strategy=SaveOnChange)
        self.assertFalse(vm.changed)
        vm.value[4] = 5
        self.assertTrue(vm.changed)

    def test_list_update(self):
        """ Tests changes on list types """
        vm = List.value_manager(List(), [1, 2, 3], strategy=SaveOnChange)
        self.assertFalse(vm.changed)
        vm.value = [4, 5, 6]
        self.assertTrue(vm.changed)

    def test_list_inplace_update(self):
        """ Tests changes on list in place types """
        vm = List.value_manager(List(), [1, 2, 3], strategy=SaveOnChange)
        self.assertFalse(vm.changed)
        vm.value.append(4)
        self.assertTrue(vm.changed)



from __future__ import unicode_literals
import logging

from mogwai._compat import float_types
from mogwai.connection import execute_query
from mogwai.exceptions import MogwaiQueryError
from element import Element, EQUAL, NOT_EQUAL, GREATER_THAN, GREATER_THAN_EQUAL, LESS_THAN, LESS_THAN_EQUAL,\
    OUT, IN, BOTH
import copy
from mogwai.properties.base import GraphProperty

logger = logging.getLogger(__name__)


class Query(object):
    """
    All query operations return a new query object, which currently deviates from blueprints.
    The blueprints query object modifies and returns the same object
    This method seems more flexible, and consistent w/ the rest of Gremlin.
    """
    _limit = None

    def __init__(self, vertex):
        self._vertex = vertex
        self._has = []
        self._interval = []
        self._labels = []
        self._direction = []
        self._vars = {}

    def count(self):
        """
        @returns: number of matching vertices
        @rtype: int
        """
        return self._execute('count', deserialize=False)

    def direction(self, direction):
        """
        @param direction: direction to compare or traverse
        @returns: Query
        """
        q = copy.copy(self)
        if self._direction:
            raise MogwaiQueryError("Direction already set")
        q._direction = direction
        return q

    def edges(self):
        """
        @returns: list[edge.Edge]
        """
        return self._execute('edges')

    def has(self, key, value, compare=EQUAL):
        """
        @param key: key to lookup
            @type key: str | mogwai.properties.GraphProperty
        @param value: value to compare
            @type value: str, float, int
        @param compare: comparison keyword
            @type compare: str
        @returns: Query
        """
        compare = "Query.Compare.{}".format(compare)

        q = copy.copy(self)
        print "Trying for key: %s with value: %s" % (key, value)
        if issubclass(type(key), property):
            logger.error("Use %s.get_property_by_name instead, this won't work" % self.__class__.__name__)
            raise MogwaiQueryError("Use %s.get_property_by_name instead, this won't work" % self.__class__.__name__)
        q._has.append((key, value, compare))
        return q

    def interval(self, key, start, end):
        """
        @returns : Query
        """
        if start > end:
            start, end = end, start

        q = copy.copy(self)
        q._interval.append((key, start, end))
        return q

    def labels(self, *args):
        """
        @param args: list of Edges
            @type args: list[edge.Edge]
        @returns: Query
        """
        tmp = []
        for x in args:
            try:
                tmp.append(x.get_label())
            except:
                tmp.append(x)

        q = copy.copy(self)
        q._labels = tmp
        return q

    def limit(self, limit):
        q = copy.copy(self)
        q._limit = limit
        return q

    def remove(self):
        """ Deletes a vertex or edge """
        return self._execute('remove', deserialize=False)

    def vertexIds(self):
        return self._execute('vertexIds', deserialize=False)

    def vertices(self):
        return self._execute('vertices')

    def _get_partial(self):
        limit = ".limit(limit)" if self._limit else ""
        dir = ".direction({})".format(self._direction) if self._direction else ""

        # do labels
        labels = ""
        if self._labels:
            labels = ["'{}'".format(x) for x in self._labels]
            labels = ", ".join(labels)
            labels = ".labels({})".format(labels)

        ### construct has clauses
        has = []

        for x in self._has:
            c = "v{}".format(len(self._vars))
            self._vars[c] = x[1]

            val = "{} as double".format(c) if isinstance(x[1], float_types) else c
            key = x[0]
            has.append("has('{}', {}, {})".format(key, val, x[2]))

        if has:
            tmp = ".".join(has)
            has = '.{}'.format(tmp)
        else:
            has = ""
        ### end construct has clauses

        intervals = []
        for x in self._interval:
            c = "v{}".format(len(self._vars))
            self._vars[c] = x[1]
            c2 = "v{}".format(len(self._vars))
            self._vars[c2] = x[2]


            val1 = "{} as double".format(c) if isinstance(x[1], float_types) else c
            val2 = "{} as double".format(c2) if isinstance(x[2], float_types) else c2

            tmp = "interval('{}', {}, {})".format(x[0], val1, val2)
            intervals.append(tmp)

        if intervals:
            intervals = ".{}".format(".".join(intervals))
        else:
            intervals = ""

        return "g.v(id).query(){}{}{}{}{}".format(labels, limit, dir, has, intervals)

    def _execute(self, func, deserialize=True):
        tmp = "{}.{}()".format(self._get_partial(), func)
        self._vars.update({"id": self._vertex._id, "limit": self._limit})
        results = execute_query(tmp, self._vars)

        if deserialize:
            return [Element.deserialize(r) for r in results]
        else:
            return results



from __future__ import unicode_literals
import datetime
from pytz import utc
from uuid import uuid4
from nose.plugins.attrib import attr

from mogwai.exceptions import MogwaiGremlinException
from mogwai.tests.base import BaseMogwaiTestCase

from mogwai.models import Vertex
from mogwai import properties
from mogwai import gremlin


class GroovyTestModel(Vertex):
    text = properties.Text()
    get_self = gremlin.GremlinMethod()
    cm_get_self = gremlin.GremlinMethod(method_name='get_self', classmethod=True)

    return_default = gremlin.GremlinValue(method_name='return_value', defaults={'val': lambda: 5000})
    return_list = gremlin.GremlinValue(property=1)
    return_value = gremlin.GremlinValue()

    arg_test1 = gremlin.GremlinValue()
    arg_test2 = gremlin.GremlinValue()


@attr('unit', 'gremlin')
class TestMethodLoading(BaseMogwaiTestCase):

    def test_method_loads_and_works(self):
        v1 = GroovyTestModel.create(text='cross fingers')

        v2 = v1.get_self()
        self.assertEqual(v1.id, v2.id)

        v3 = v1.cm_get_self(v1.id)
        self.assertEqual(v1.id, v3.id)

        v1.delete()


@attr('unit', 'gremlin', 'gremlin2')
class TestMethodArgumentHandling(BaseMogwaiTestCase):

    def test_callable_defaults(self):
        """
        Tests that callable default arguments are called
        """
        v1 = GroovyTestModel.create(text='cross fingers')
        self.assertEqual(v1.return_default(), 5000)
        v1.delete()

    def test_gremlin_value_enforces_single_object_returned(self):
        """
        Tests that a GremlinValue instance raises an error if more than one object is returned
        """
        v1 = GroovyTestModel.create(text='cross fingers')
        with self.assertRaises(MogwaiGremlinException):
            v1.return_list
        v1.delete()

    def test_type_conversion(self):
        """ Tests that the gremlin method converts certain python objects to their gremlin equivalents """
        v1 = GroovyTestModel.create(text='cross fingers')

        now = datetime.datetime.now(tz=utc)
        self.assertEqual(v1.return_value(now), properties.DateTime().to_database(now))

        uu = uuid4()
        self.assertEqual(v1.return_value(uu), properties.UUID().to_database(uu))
        v1.delete()

    def test_initial_arg_name_isnt_set(self):
        """ Tests that the name of the first argument in a instance method """
        v = GroovyTestModel.create(text='cross fingers')

        self.assertEqual(v, v.arg_test1())
        self.assertEqual(v, v.arg_test2())

        v.delete()



from __future__ import unicode_literals
from mogwai.tests import BaseMogwaiTestCase
from mogwai.properties import GraphProperty
from nose.plugins.attrib import attr
from mogwai.exceptions import ValidationError
from mogwai._compat import print_


@attr('unit', 'property')
class GraphPropertyBaseClassTestCase(BaseMogwaiTestCase):
    """ Test Base Strategy Callable Object """
    klass = GraphProperty
    good_cases = ('', 'a', 1, 1.1, None, [], [1], {}, {'a': 1})
    bad_cases = ()

    def test_subclass(self):
        """ Test if Property is a GraphProperty """
        self.assertIsSubclass(self.klass, GraphProperty)

    def test_validation(self):
        for case in self.good_cases:
            print_("testing good case: %s" % (case, ))
            self.assertNotRaise(self.klass().validate, case)

        for case in self.bad_cases:
            print_("testing bad case: %s" % (case, ))
            self.assertRaises(ValidationError, self.klass().validate, case)



from __future__ import unicode_literals
from nose.plugins.attrib import attr
from base_tests import GraphPropertyBaseClassTestCase
from mogwai.properties.properties import Slug
from mogwai.models import Vertex
from mogwai._compat import print_


@attr('unit', 'property', 'property_slug')
class SlugPropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = Slug
    good_cases = ('ab', '12', 'ab12', '12ab', None)
    bad_cases = ('@', 'a!', '12!', 'ab12!', 0, 1.1, [], (), {})

    def test_slug_defaults(self):
        p = Slug(default='ab')


class SlugTestVertex(Vertex):
    element_type = 'test_slug_vertex'

    test_val = Slug()


@attr('unit', 'property', 'property_slug')
class SlugVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_slug_io(self):
        print_("creating vertex")
        dt = SlugTestVertex.create(test_val='ab12')
        print_("getting vertex from vertex: %s" % dt)
        dt2 = SlugTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = SlugTestVertex.create(test_val='12ab')
        print_("\ncreated vertex: %s" % dt)
        dt2 = SlugTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, '12ab')
        print_("deleting vertex")
        dt2.delete()


from __future__ import unicode_literals
from nose.plugins.attrib import attr
from base_tests import GraphPropertyBaseClassTestCase
from mogwai.properties.properties import URL
from mogwai.models import Vertex
from mogwai.exceptions import ValidationError
from mogwai._compat import print_


@attr('unit', 'property', 'property_url')
class URLPropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = URL
    good_cases = ('http://subdomain.domain.com/',
                  'https://domain.com/',
                  'http://domain.com/path/',
                  'http://domain.com/path/?params=1',
                  'http://subdomain.com/path/deep/?params=2&second=3')
    bad_cases = ('htp://fail.com',
                 'fail.com/path/',
                 '/asdf/')

    def test_url_edge_cases(self):
        p = URL(default='http://www.abc.com', max_length=20, min_length=17)
        with self.assertRaises(ValidationError):
            p.validate('http://www.s.com')
        with self.assertRaises(ValidationError):
            p.validate('http://www.thisiswaytoolongggg.com')


class URLTestVertex(Vertex):
    element_type = 'test_url_vertex'

    test_val = URL(required=True)


@attr('unit', 'property', 'property_url')
class URLVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_url_io(self):
        print_("creating vertex")
        dt = URLTestVertex.create(test_val='http://wellaware.us')
        print_("getting vertex from vertex: %s" % dt)
        dt2 = URLTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = URLTestVertex.create(test_val='http://www.wellaware.us/')
        print_("\ncreated vertex: %s" % dt)
        dt2 = URLTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, 'http://www.wellaware.us/')
        print_("deleting vertex")
        dt2.delete()


from __future__ import unicode_literals
import inspect
import logging

from mogwai._compat import array_types, string_types
from mogwai.connection import execute_query
from mogwai.exceptions import MogwaiException, ElementDefinitionException, MogwaiQueryError
from mogwai.gremlin import GremlinMethod
from element import Element, ElementMetaClass, vertex_types

logger = logging.getLogger(__name__)


class VertexMetaClass(ElementMetaClass):
    """Metaclass for vertices."""

    def __new__(mcs, name, bases, body):

        #short circuit element_type inheritance
        body['element_type'] = body.pop('element_type', None)

        klass = super(VertexMetaClass, mcs).__new__(mcs, name, bases, body)

        if not klass.__abstract__:
            element_type = klass.get_element_type()
            if element_type in vertex_types and str(vertex_types[element_type]) != str(klass):
                logger.debug(ElementDefinitionException("%s is already registered as a vertex: \n\tmcs: %s\n\tname: %s\n\tbases: %s\n\tbody: %s" % (element_type, mcs, name, bases, body)))
            else:
                vertex_types[element_type] = klass

            ##index requested indexed columns
            #klass._create_indices()

        return klass


class EnumVertexBaseMeta(VertexMetaClass):
    """
    This metaclass allows you to access MyVertexModel as if it were an enum.
    Ex. MyVertexModel.FOO

    The values are cached in a dictionary. This is useful if the number of MyVertexModels is small,
    however it it grows too large, you should be doing it a different way.

    This looks for a special (optional) function named `enum_generator` in your model and calls that
    to generate the ENUM for the model.

    There is an additional optional model attribute that can be set `__enum_id_only__` (defaults to True)
    which dictates whether or not just the Vertex ID is stored, or the whole Vertex in cache.
    """

    enums = None

    def __getattr__(cls, key):
        # property name to use for keying for the enum
        # method for handling name mangling, default to passthrough mode which subs spaces for underscores and caps
        store_model = getattr(cls, '__enum_id_only__', True)

        def get_enum_keyword(enum):
            return getattr(enum, 'enum_generator', lambda: (getattr(enum, 'name', '').replace(' ', '_').upper()))()

        if key.isupper():
            if cls.enums is None:
                cls.enums = dict(
                    [(get_enum_keyword(enum), enum._id if store_model else enum) for enum in cls.all()]
                )
            id = cls.enums.get(key, None)
            if not id:
                # make one attempt to load any new models
                cls.enums = dict(
                    [(get_enum_keyword(enum), enum._id if store_model else enum) for enum in cls.all()]
                )
                id = cls.enums.get(key, None)
                if not id:
                    raise AttributeError(key)
            return id
        else:
            return super(EnumVertexBaseMeta, cls).__getattr__(key)


class Vertex(Element):
    """ The Vertex model base class.

    The element type is auto-generated from the subclass name, but can optionally be set manually
    """
    __metaclass__ = VertexMetaClass
    __abstract__ = True

    gremlin_path = 'vertex.groovy'

    _save_vertex = GremlinMethod()
    _traversal = GremlinMethod()
    _delete_related = GremlinMethod()
    find_by_value = GremlinMethod(classmethod=True)

    element_type = None

    FACTORY_CLASS = None

    def __repr__(self):
        return "%s(element_type=%s, id=%s, values=%s)" % (self.__class__.__name__,
                                                          self.element_type,
                                                          getattr(self, '_id', None),
                                                          getattr(self, '_values', []))

    @classmethod
    def get_element_type(cls):
        """
        Returns the element type for this vertex.

        @returns: str

        """
        return cls._type_name(cls.element_type)

    @classmethod
    def all(cls, ids=[], as_dict=False, match_length=True):
        """
        Load all vertices with the given ids from the graph. By default this will return a list of vertices but if
        as_dict is True then it will return a dictionary containing ids as keys and vertices found as values.

        @param ids: A list of titan ids
            @type ids: list
        @param as_dict: Toggle whether to return a dictionary or list
            @type as_dict: boolean
        @returns: dict | list

        """
        if not isinstance(ids, array_types):
            raise MogwaiQueryError("ids must be of type list or tuple")

        if len(ids) == 0:
            results = execute_query('g.V("element_type","%s").toList()' % cls.get_element_type())

        else:
            strids = [str(i) for i in ids]

            results = execute_query('ids.collect{g.v(it)}', {'ids': strids})
            results = filter(None, results)

            if len(results) != len(ids) and match_length:
                raise MogwaiQueryError("the number of results don't match the number of ids requested")

        objects = []
        for r in results:
            try:
                objects += [Element.deserialize(r)]
            except KeyError:  # pragma: no cover
                raise MogwaiQueryError('Vertex type "%s" is unknown' % r.get('element_type', ''))

        if as_dict:  # pragma: no cover
            return {v._id: v for v in objects}

        return objects

    def _reload_values(self):
        """
        Method for reloading the current vertex by reading its current values from the database.
        """
        reloaded_values = {}
        results = execute_query('g.v(id)', {'id': self._id})
        #del results['_id']
        del results['_type']
        reloaded_values['_id'] = results['_id']
        for name, value in results.get('_properties', {}).items():
            reloaded_values[name] = value
        return reloaded_values

    @classmethod
    def get(cls, id):
        """
        Look up vertex by its ID. Raises a DoesNotExist exception if a vertex with the given vid was not found.
        Raises a MultipleObjectsReturned exception if the vid corresponds to more than one vertex in the graph.

        @param id: The ID of the vertex
            @type id: str
        @returns: mogwai.models.Vertex

        """
        try:
            results = cls.all([id])
            if len(results) > 1:  # pragma: no cover
                # This requires to titan to be broken.
                raise cls.MultipleObjectsReturned

            result = results[0]
            if not isinstance(result, cls):
                raise cls.WrongElementType(
                    '%s is not an instance or subclass of %s' % (result.__class__.__name__, cls.__name__)
                )
            return result
        except MogwaiQueryError as e:
            logger.exception(e)
            raise cls.DoesNotExist

    def save(self, *args, **kwargs):
        """
        Save the current vertex using the configured save strategy, the default save strategy is to re-save all
        fields every time the object is saved.
        """
        super(Vertex, self).save(*args, **kwargs)
        params = self.as_save_params()
        params['element_type'] = self.get_element_type()
        result = self._save_vertex(params)
        self._id = result._id
        for k, v in self._values.items():
            v.previous_value = result._values[k].previous_value
        return result

    def delete(self):
        """ Delete the current vertex from the graph. """
        if self.__abstract__:
            raise MogwaiQueryError('Cant delete abstract elements')
        if self._id is None:  # pragma: no cover
            return self
        query = """
        g.removeVertex(g.v(id))
        g.stopTransaction(SUCCESS)
        """
        results = execute_query(query, {'id': self._id})

    def _simple_traversal(self,
                          operation,
                          labels,
                          limit=None,
                          offset=None,
                          types=None):
        """
        Perform simple graph database traversals with ubiquitous pagination.

        @param operation: The operation to be performed
            @type operation: str
        @param labels: The edge labels to be used
            @type labels: list of Edges or strings
        @param start: The starting offset
            @type start: int
        @param max_results: The maximum number of results to return
            @type max_results: int
        @param types: The list of allowed result elements
            @type types: list

        """
        from edge import Edge

        label_strings = []
        for label in labels:
            if inspect.isclass(label) and issubclass(label, Edge):
                label_string = label.get_label()
            elif isinstance(label, Edge):
                label_string = label.get_label()
            elif isinstance(label, string_types):
                label_string = label
            else:
                raise MogwaiException('traversal labels must be edge classes, instances, or strings')
            label_strings.append(label_string)

        allowed_elts = None
        if types is not None:
            allowed_elts = []
            for e in types:
                if issubclass(e, Vertex):
                    allowed_elts += [e.get_element_type()]
                elif issubclass(e, Edge):
                    allowed_elts += [e.get_label()]

        if limit is not None and offset is not None:
            start = offset
            end = offset + limit
        else:
            start = end = None

        return self._traversal(operation,
                               label_strings,
                               start,
                               end,
                               allowed_elts)

    def _simple_deletion(self, operation, labels):
        """
        Perform simple bulk graph deletion operation.

        @param operation: The operation to be performed
            @type operation: str
        @param label: The edge label to be used
            @type label: str or Edge

        """
        from edge import Edge

        label_strings = []
        for label in labels:
            if inspect.isclass(label) and issubclass(label, Edge):
                label_string = label.get_label()
            elif isinstance(label, Edge):
                label_string = label.get_label()
            elif isinstance(label, string_types):
                label_string = label
            else:
                raise MogwaiException('traversal labels must be edge classes, instances, or strings')
            label_strings.append(label_string)

        return self._delete_related(operation, label_strings)

    def outV(self, *labels, **kwargs):
        """
        Return a list of vertices reached by traversing the outgoing edge with the given label.

        @param labels: pass in the labels to follow in as positional arguments
            @type labels: str or BaseEdge
        @param limit: The number of the page to start returning results at
            @type limit: int or None
        @param offset: The maximum number of results to return
            @type offset: int or None
        @param types: A list of allowed element types
            @type types: list

        """
        return self._simple_traversal('outV', labels, **kwargs)

    def inV(self, *labels, **kwargs):
        """
        Return a list of vertices reached by traversing the incoming edge with the given label.

        @param label: The edge label to be traversed
            @type label: str or BaseEdge
        @param limit: The number of the page to start returning results at
            @type limit: int or None
        @param offset: The maximum number of results to return
            @type offset: int or None
        @param types: A list of allowed element types
            @type types: list

        """
        return self._simple_traversal('inV', labels, **kwargs)

    def outE(self, *labels, **kwargs):
        """
        Return a list of edges with the given label going out of this vertex.

        @param label: The edge label to be traversed
            @type label: str or BaseEdge
        @param limit: The number of the page to start returning results at
            @type limit: int or None
        @param offset: The maximum number of results to return
            @type offset: int or None
        @param types: A list of allowed element types
            @type types: list

        """
        return self._simple_traversal('outE', labels, **kwargs)

    def inE(self, *labels, **kwargs):
        """
        Return a list of edges with the given label coming into this vertex.

        @param label: The edge label to be traversed
            @type label: str or BaseEdge
        @param limit: The number of the page to start returning results at
            @type limit: int or None
        @param offset: The maximum number of results to return
            @type offset: int or None
        @param types: A list of allowed element types
            @type types: list

        """
        return self._simple_traversal('inE', labels, **kwargs)

    def bothE(self, *labels, **kwargs):
        """
        Return a list of edges both incoming and outgoing from this vertex.

        @param label: The edge label to be traversed (optional)
            @type label: str or BaseEdge or None
        @param limit: The number of the page to start returning results at
            @type limit: int or None
        @param offset: The maximum number of results to return
            @type offset: int or None
        @param types: A list of allowed element types
            @type types: list

        """
        return self._simple_traversal('bothE', labels, **kwargs)

    def bothV(self, *labels, **kwargs):
        """
        Return a list of vertices both incoming and outgoing from this vertex.

        @param label: The edge label to be traversed (optional)
            @type label: str or BaseEdge or None
        @param limit: The number of the page to start returning results at
            @type limit: int or None
        @param offset: The maximum number of results to return
            @type offset: int or None
        @param types: A list of allowed element types
            @type types: list

        """
        return self._simple_traversal('bothV', labels, **kwargs)

    def delete_outE(self, *labels):
        """Delete all outgoing edges with the given label."""
        self._simple_deletion('outE', labels)

    def delete_inE(self, *labels):
        """Delete all incoming edges with the given label."""
        self._simple_deletion('inE', labels)

    def delete_outV(self, *labels):
        """Delete all outgoing vertices connected with edges with the given label."""
        self._simple_deletion('outV', labels)

    def delete_inV(self, *labels):
        """Delete all incoming vertices connected with edges with the given label."""
        self._simple_deletion('inV', labels)

    def query(self):
        from query import Query
        return Query(self)


import collections
import pyparsing
import re


# Cache of parsed files
_parsed_file_cache = {}


class GroovyFunctionParser(object):
    """
    Given a string containing a single function definition this class will parse the function definition and return
    information regarding it.
    """

    # Simple Groovy sub-grammar definitions
    KeywordDef = pyparsing.Keyword('def')
    VarName = pyparsing.Regex(r'[A-Za-z_]\w*')
    FuncName = VarName
    FuncDefn = KeywordDef + FuncName + "(" + pyparsing.delimitedList(VarName) + ")" + "{"

    # Result named tuple
    GroovyFunction = collections.namedtuple('GroovyFunction', ['name', 'args', 'body', 'defn'])

    @classmethod
    def parse(cls, data):
        """
        Parse the given function definition and return information regarding the contained definition.

        @param data: The function definition in a string
            @type data: str
        @returns: dict

        """
        try:
            # Parse the function here
            result = cls.FuncDefn.parseString(data)
            result_list = result.asList()
            args = result_list[3:result_list.index(')')]
            # Return single line or multi-line function body
            fn_body = re.sub(r'[^\{]+\{', '', data, count=1)
            parts = fn_body.strip().split('\n')
            fn_body = '\n'.join(parts[0:-1])
            return cls.GroovyFunction(result[1], args, fn_body, data)
        except Exception as ex:
            return {}


def parse(file):
    """
    Parse Groovy code in the given file and return a list of information about each function necessary for usage in
    queries to database.

    @param file: The file containing groovy code.
        @type file: str
    @returns: list

    """
    # Check cache before parsing file
    global _parsed_file_cache
    if file in _parsed_file_cache:
        return _parsed_file_cache[file]

    FuncDefnRegexp = r'^def.*\{'
    FuncEndRegexp = r'^\}.*$'
    with open(file, 'r') as f:
        data = f.read()
    file_lines = data.split("\n")
    all_fns = []
    fn_lines = ''
    for line in file_lines:
        if len(fn_lines) > 0:
            if re.match(FuncEndRegexp, line):
                fn_lines += line + "\n"
                all_fns.append(fn_lines)
                fn_lines = ''
            else:
                fn_lines += line + "\n"
        elif re.match(FuncDefnRegexp, line):
            fn_lines += line + "\n"

    func_results = []
    for fn in all_fns:
        func_results += [GroovyFunctionParser.parse(fn)]

    _parsed_file_cache[file] = func_results
    return func_results



from __future__ import unicode_literals
from nose.plugins.attrib import attr
from base_tests import GraphPropertyBaseClassTestCase
from mogwai.properties.properties import List
from mogwai.models import Vertex
from mogwai._compat import print_


@attr('unit', 'property', 'property_list')
class ListPropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = List
    good_cases = ([], ['val'], (), tuple('val'))
    bad_cases = (0, 1.1, None, 'val', {}, {'value': 1})


class ListTestVertex(Vertex):
    element_type = 'test_list_vertex'

    test_val = List()


@attr('unit', 'property', 'property_list')
class ListVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_list_io(self):
        print_("creating vertex")
        dt = ListTestVertex.create(test_val=['a', 'b'])
        print_("getting vertex from vertex: %s" % dt)
        dt2 = ListTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = ListTestVertex.create(test_val=[1, 2])
        print_("\ncreated vertex: %s" % dt)
        dt2 = ListTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, [1, 2])
        print_("deleting vertex")
        dt2.delete()


from __future__ import unicode_literals
import os
from nose.plugins.attrib import attr
from mogwai.tests.base import BaseMogwaiTestCase
from mogwai.gremlin import parse


@attr('unit', 'groovy')
class GroovyScannerTest(BaseMogwaiTestCase):
    """
    Test Groovy language scanner
    """

    def test_parsing_complicated_function(self):
        """ Groovy test parser functionality """
        groovy_file = os.path.join(os.path.dirname(__file__), 'groovy_test_model.groovy')
        result = parse(groovy_file)
        self.assertEqual(len(result[6].body.split('\n')), 8)

        result_map = {x.name: x for x in result}
        self.assertIn('get_self', result_map)
        self.assertIn('return_value', result_map)
        self.assertIn('long_func', result_map)


from __future__ import unicode_literals
from mogwai._compat import PY2
from mogwai.tests import BaseMogwaiTestCase, testcase_docstring_sub
from mogwai.properties.validators import *
from nose.plugins.attrib import attr
from decimal import Decimal as _D
import datetime
from pytz import utc


@attr('unit', 'validators')
class ValidatorBaseClassTestCase(BaseMogwaiTestCase):
    """ Test Base Validator Callables """
    klass = BaseValidator()
    good_cases = ('', 'a', 1, 1.1, None, [], [1], {}, {'a': 1})
    bad_cases = ()

    def test_callable(self):
        """ Test that the validator is callable """
        self.assertTrue(issubclass(self.klass.__class__, BaseValidator))
        # test validator is callable
        self.assertTrue(callable(self.klass))

    def test_good_validation(self):
        for case in self.good_cases:
            print "testing case: %s" % (case, )
            self.assertNotRaise(self.klass, case)

    def test_bad_validation(self):
        for case in self.bad_cases:
            print "testing case: %s" % (case, )
            self.assertRaises(ValidationError, self.klass, case)


@attr('unit', 'validators')
class BooleanValidatorTestCase(ValidatorBaseClassTestCase):
    """ Boolean Validator """
    klass = BooleanValidator()
    good_cases = (True, False)
    bad_cases = (0, 1.1, 'val', [], (), {}, None)


@attr('unit', 'validators')
class NumericValidatorTestCase(ValidatorBaseClassTestCase):
    klass = NumericValidator()
    if PY2:
        good_cases = (0, 1.1, long(1), True, False)
    else:
        good_cases = (0, 1.1, True, False)
    bad_cases = ('val', [], (), {}, None)


@attr('unit', 'validators')
class FloatValidatorTestCase(ValidatorBaseClassTestCase):
    klass = FloatValidator()
    good_cases = (1.1, )
    if PY2:
        bad_cases = ('val', [], (), {}, None, 0, long(1), False)
    else:
        bad_cases = ('val', [], (), {}, None, 0, False)


@attr('unit', 'validators')
class DecimalValidatorTestCase(FloatValidatorTestCase):
    klass = DecimalValidator()
    good_cases = (1.1, _D(1.1))


@attr('unit', 'validators')
class IntegerValidatorTestCase(ValidatorBaseClassTestCase):
    klass = IntegerValidator()
    if PY2:
        good_cases = (1, long(1), False, True)
    else:
        good_cases = (1, False, True)
    bad_cases = ('val', [], (), {}, None, 0.1)


@attr('unit', 'validators')
class LongValidatorTestCase(ValidatorBaseClassTestCase):
    klass = LongValidator()
    if PY2:
        good_cases = (1, long(2), True, False)
    else:
        good_cases = (1, True, False)
    bad_cases = ('val', [], (), {}, None)


@attr('unit', 'validators')
class PositiveIntegerValidatorTestCase(ValidatorBaseClassTestCase):
    klass = PositiveIntegerValidator()
    if PY2:
        good_cases = (1, long(2), True, False)
        bad_cases = ('val', [], (), {}, None, -1, long(-2))
    else:
        good_cases = (1, True, False)
        bad_cases = ('val', [], (), {}, None, -1)


@attr('unit', 'validators')
class StringValidatorTestCase(ValidatorBaseClassTestCase):
    klass = StringValidator()
    good_cases = ('val', )
    if PY2:
        bad_cases = (1.1, [], (), {}, None, 0, long(1), False)
    else:
        bad_cases = (1.1, [], (), {}, None, 0, False)


@attr('unit', 'validators')
class ListValidatorTestCase(ValidatorBaseClassTestCase):
    klass = ListValidator()
    good_cases = ([], ['val'], (), tuple('val'))
    if PY2:
        bad_cases = ('val', {}, None, 0, long(1), False)
    else:
        bad_cases = ('val', {}, None, 0, False)


@attr('unit', 'validators')
class DictValidatorTestCase(ValidatorBaseClassTestCase):
    klass = DictValidator()
    good_cases = ({}, {'value': 1})
    if PY2:
        bad_cases = ('val', [], (), None, 0, long(1), False, 1.1)
    else:
        bad_cases = ('val', [], (), None, 0, False, 1.1)


@attr('unit', 'validators')
class DateTimeValidatorTestCase(ValidatorBaseClassTestCase):
    klass = DateTimeValidator()
    good_cases = (datetime.datetime.now(), None)
    if PY2:
        bad_cases = ('val', [], (), {}, 0, long(1), False, 1.1)
    else:
        bad_cases = ('val', [], (), {}, 0, False, 1.1)


@attr('unit', 'property')
class DateTimeUTCValidatorTestCase(ValidatorBaseClassTestCase):
    klass = DateTimeUTCValidator()
    good_cases = (datetime.datetime.now(tz=utc), None)
    if PY2:
        bad_cases = ('val', [], (), {}, 0, long(1), False, 1.1, datetime.datetime.now())
    else:
        bad_cases = ('val', [], (), {}, 0, False, 1.1, datetime.datetime.now())


@attr('unit', 'validators')
class URLValidatorTestCase(ValidatorBaseClassTestCase):
    """ URL Validator """

    klass = URLValidator()
    good_cases = ('http://subdomain.domain.com/',
                  'https://domain.com/',
                  'http://domain.com/path/',
                  'http://domain.com/path/?params=1',
                  'http://subdomain.com/path/deep/?params=2&second=3')
    bad_cases = ('htp://fail.com',
                 'fail.com/path/',
                 '/asdf/')


@attr('unit', 'validators')
class EmailValidatorTestCase(ValidatorBaseClassTestCase):
    """ Email Validator """

    klass = EmailValidator()
    good_cases = ('test@test.com', 'test-alice@subdomain.domain.com')
    bad_cases = ('', 'bob', '@test.com', 'bob@test')


@attr('unit', 'validators')
class SlugValidatorTestCase(ValidatorBaseClassTestCase):
    """ Slug Validator """

    klass = validate_slug
    good_cases = ('ab', '12', 'ab12', '12ab')
    bad_cases = ('@', 'a!', '12!', 'ab12!')


@attr('unit', 'validators')
class IPV4ValidatorTestCase(ValidatorBaseClassTestCase):
    """ IPV4 Validator """

    klass = validate_ipv4_address
    good_cases = ('1.2.3.4', '0.0.0.0', '255.255.255.255')
    bad_cases = ('0', '0.', '0.0', '0.0.', '0.0.0', '0.0.0.', '256.256.256.256', '1.2.3.256')


@attr('unit', 'validators')
class IPV6ValidatorTestCase(ValidatorBaseClassTestCase):
    """ IPV6 Validator """

    klass = validate_ipv6_address
    good_cases = ('1:2:3:4:5:6:7:8', '1::', '1:2:3:4:5:6:7::',
                  '1::8', '1:2:3:4:5:6::8', '1:2:3:4:5:6::8',
                  '1::7:8', '1:2:3:4:5::7:8', '1:2:3:4:5::8',
                  '1::6:7:8', '1:2:3:4::6:7:8', '1:2:3:4::8',
                  '1::5:6:7:8', '1:2:3::5:6:7:8', '1:2:3::8',
                  '1::4:5:6:7:8', '1:2::4:5:6:7:8', '1:2::8',
                  '1::3:4:5:6:7:8', '1::3:4:5:6:7:8', '1::8',
                  '::2:3:4:5:6:7:8', '::2:3:4:5:6:7:8', '::8', '::')
    bad_cases = ('0', '0.', '0.0', '0.0.', '0.0.0', '0.0.0.', '256.256.256.256', '1.2.3.256')


@attr('unit', 'validators')
class IPV6With4ValidatorTestCase(ValidatorBaseClassTestCase):
    """ IPv4 mapped/translated/embedded IPv6 Validator """

    klass = validate_ipv6_ipv4_address
    good_cases = ('1:2:3:4:5:6:7:8', '1::', '1:2:3:4:5:6:7::',  # IPv6
                  '1::8', '1:2:3:4:5:6::8', '1:2:3:4:5:6::8',
                  '1::7:8', '1:2:3:4:5::7:8', '1:2:3:4:5::8',
                  '1::6:7:8', '1:2:3:4::6:7:8', '1:2:3:4::8',
                  '1::5:6:7:8', '1:2:3::5:6:7:8', '1:2:3::8',
                  '1::4:5:6:7:8', '1:2::4:5:6:7:8', '1:2::8',
                  '1::3:4:5:6:7:8', '1::3:4:5:6:7:8', '1::8',
                  '::2:3:4:5:6:7:8', '::2:3:4:5:6:7:8', '::8', '::',
                  '::255.255.255.255', '::ffff:255.255.255.255', '::ffff:0:255.255.255.255',  # Mapped/Translated
                  '2001:db8:3:4::192.0.2.33', '64:ff9b::192.0.2.33',  # Embedded
                  )
    bad_cases = ('0', '0.', '0.0', '0.0.', '0.0.0', '0.0.0.', '256.256.256.256', '1.2.3.256', '1.2.3.4')


@attr('unit', 'validators')
class UUID4ValidatorTestCase(ValidatorBaseClassTestCase):
    klass = validate_uuid4
    good_cases = ('bb19eaed-c946-4cef-8001-7cc3357cc439', )
    if PY2:
        bad_cases = ('val', [], (), {}, None, 0, long(1), False, 1.1)
    else:
        bad_cases = ('val', [], (), {}, None, 0, False, 1.1)


@attr('unit', 'validators')
class UUID1ValidatorTestCase(UUID4ValidatorTestCase):
    klass = validate_uuid1


from __future__ import unicode_literals
import copy
import warnings

from mogwai.exceptions import ValidationError, MogwaiException
from strategy import Strategy, SaveAlways, SaveOnce
from validators import pass_all_validator


class BaseValueManager(object):
    """
    Value managers are used to manage values pulled from the database and track state
    changes.

    These are useful for save strategies.
    """

    def __init__(self, graph_property, value, strategy=SaveAlways):
        """
        Initialize the value manager.

        @param graph_property: The graph property to manage
            @type graph_property: mogwai.properties.base.GraphProperty
        @param value: The initial value of the column
            @type value: Object
        @param strategy: The callable condition to compare against. If none given, it won't be used
            @type strategy: mogwai.properties.strategy.Strategy
        """
        self._create_private_fields()

        self.graph_property = graph_property
        self._previous_value = copy.copy(value)
        self.value = value
        self.strategy = strategy
        if not issubclass(self.strategy.__class__, Strategy):
            warnings.warn("Condition is not derived from mogwai.properties.strategy.Strategy and will be ignored!",
                          category=SyntaxWarning)

    def __repr__(self):
        return "%s(property=%s, value=%s, previous_value=%s, strategy=%s)" % (self.__class__.__name__,
                                                                              self.graph_property,
                                                                              self.value,
                                                                              self.previous_value,
                                                                              self.strategy)

    def _create_private_fields(self):
        self._previous_value = None

    @property
    def previous_value(self):
        return self._previous_value

    @previous_value.setter
    def previous_value(self, val):
        self._previous_value = copy.copy(val)

    @property
    def deleted(self):
        """
        Indicates whether or not this value has been deleted.

        @returns: bool

        """
        return self.value is None and self.previous_value is not None

    @property
    def changed(self):
        """
        Indicates whether or not this value has changed.

        @returns: bool
        """
        try:
            return self.strategy.condition(self.previous_value,
                                           self.value,
                                           has_changed=(self.value != self.previous_value),
                                           graph_property=self.graph_property)
        except:
            return self.value != self.previous_value

    def getval(self):
        """Return the current value."""
        return self.value

    def setval(self, val):
        """
        Updates the current value.

        @param val: The new value
        @type val: mixed

        """
        self.value = val

    def delval(self):
        """Delete a given value"""
        self.value = None

    def get_property(self):
        """
        Returns a value-managed property attributes

        :rtype: property

        """
        _get = lambda slf: self.getval()
        _set = lambda slf, val: self.setval(val)
        _del = lambda slf: self.delval()

        if self.graph_property.can_delete:
            return property(_get, _set, _del)
        else:
            return property(_get, _set)


class GraphProperty(object):
    """Base class for graph property types"""
    data_type = "Object"
    value_manager = BaseValueManager
    validator = pass_all_validator
    instance_counter = 0

    def __init__(self, description=None, primary_key=False, index=False, index_ext=None, db_field=None, choices=None,
                 default=None, required=False, save_strategy=SaveAlways, unique=None, db_field_prefix=''):
        """
        Initialize this graph property with the given information.

        @param description: description of this field
            @type description: basestring | str
        @param primary_key: Indicates whether or not this is primary key
            @type primary_key: bool
        @param index: Indicates whether or not this field is indexed
            @type index: bool
        @param db_field: The property this field will map to in the database
            @type db_field: basestring | str
        @param choices: A dict of possible choices where the key is the value to store, and the value is the
                        user-friendly value
            @type choices: tuple(tuple(Object, Object)) | list[list[Object, Object]] | None
        @param default: Value or callable with no args to set default value
            @type default: Callable | Number
        @param required: Whether or not this field is required
            @type required: bool
        @param save_strategy: Strategy used when saving the value of the column
            @type save_strategy: strategy.Strategy
        @param unique: Uniqueness constraint left in for backwards compatibility -- used by Spec system.
            @type unique: bool
        @param db_field_prefix: The property prefix associated with the Model.
            @type db_field_prefix: basestring | None

        """
        self.description = description
        self.primary_key = primary_key
        self.index = index
        self.index_ext = index_ext
        self.db_field = db_field
        self.db_field_prefix = db_field_prefix
        self.default = default
        self.required = required
        self.save_strategy = save_strategy
        self.unique = unique
        self.choices = choices

        #the graph property name in the model definition
        self.property_name = None

        #self.value = None

        #keep track of instantiation order
        self.position = GraphProperty.instance_counter
        GraphProperty.instance_counter += 1

    def __repr__(self):
        return "%s(name=%s, pos=%s, default=%s, db_field_name=%s)" % (self.__class__.__name__,
                                                                      self.property_name,
                                                                      self.position,
                                                                      self.default,
                                                                      self.db_field_name)

    @classmethod
    def get_value_from_choices(cls, value, choices):
        """ Returns the key for the choices tuple of tuples

        Note if you are using classes, they must implement the __in__ and __eq__ for the logical comparison.

        @param value: The raw value to test if it exists in the valid choices. Could be the key or the value in the dict
            @type value: Object
        @returns Object
        """
        if not choices:
            return None
        for key, val in choices:
            if value in (key, val):
                return key
        return None

    def validate(self, value):
        """
        Returns a cleaned and validated value. Raises a ValidationError if there's a problem

        @returns: Object
        """
        if self.choices:
            orig_value = value
            value = GraphProperty.get_value_from_choices(value, self.choices)
            if value is None:
                raise ValidationError('%s - Value `%s` is not in available choices' % (self.db_field_name, orig_value))
        if value is None:
            if self.has_default:
                return self.validator(self.get_default())
            elif self.required:
                raise ValidationError('%s - None values are not allowed' % self.db_field_name)
        return self.validator(value)

    def to_python(self, value):
        """
        Converts data from the database into python values raises a ValidationError if the value can't be converted

        @returns: Object
        """
        return value

    def to_database(self, value):
        """
        Converts python value into database value

        @returns: Object
        """
        if value is None and self.has_default:
            return self.get_default()
        return value

    @property
    def has_default(self):
        """
        Indicates whether or not this graph property has a default value.

        @returns: bool
        """
        return self.default is not None

    @property
    def can_delete(self):
        return not self.primary_key

    def should_save(self, first_save=False):
        """
        Indicates whether or not the property should be saved based on it's save strategy.

        @returns: bool
        """
        return self.get_save_strategy().condition(previous_value=self.value_manager.previous_value,
                                                  value=self.value_manager.value,
                                                  has_changed=self.value_manager.changed,
                                                  first_save=first_save,
                                                  graph_property=self)

    def get_save_strategy(self):
        """
        Returns the save strategy attached to this graph property.

        @returns: Callable

        """
        return self.save_strategy or (SaveAlways if not self.primary_key else SaveOnce)

    def get_default(self):
        """
        Returns the default value for this graph property if one is available.

        @returns: Object | None
        """
        if self.has_default:
            if callable(self.default):
                return self.default()
            else:
                return self.default

    def set_property_name(self, name):
        """
        Sets the graph property name during document class construction.

        This value will be ignored if db_field is set in __init__

        @param name: The name of this graph property
            @type name: str
        """
        self.property_name = name

    def set_db_field_prefix(self, prefix, override=False):
        """
        Sets the graph property name prefix during document class construction.
        """
        if not override and not self.db_field_prefix:
            self.db_field_prefix = prefix.rstrip('_') + '_'
            if self.db_field_prefix == '_':
                self.db_field_prefix = ''

    @property
    def has_db_field_prefix(self):
        """
        Determines if a field prefix has already been defined.
        """
        if self.db_field_prefix not in [None, '']:
            return True
        return False

    @property
    def db_field_name(self):
        """
        Returns the name of the mogwai name of this graph property

        @returns: basestring | str
        """
        return (self.db_field_prefix or '') + (self.db_field or self.property_name or '')


import logging
from mogwai._compat import array_types, integer_types, string_types
from mogwai.connection import execute_query
from mogwai.exceptions import ElementDefinitionException, MogwaiQueryError, ValidationError
from mogwai.gremlin import GremlinMethod
from element import Element, ElementMetaClass, edge_types

logger = logging.getLogger(__name__)


class EdgeMetaClass(ElementMetaClass):
    """Metaclass for edges."""

    def __new__(mcs, name, bases, body):
        #short circuit element_type inheritance
        body['label'] = body.pop('label', None)

        klass = super(EdgeMetaClass, mcs).__new__(mcs, name, bases, body)

        if not klass.__abstract__:
            label = klass.get_label()
            if label in edge_types and str(edge_types[label]) != str(klass):
                # Catch imports from other modules, don't reload module into vertex types, only load once
                logger.debug(ElementDefinitionException("%s is already registered as a edge: \n\tmcs: %s\n\tname: %s\n\tbases: %s\n\tbody: %s" % (label, mcs, name, bases, body)))
            else:
                edge_types[klass.get_label()] = klass
        return klass


class Edge(Element):
    """Base class for all edges."""

    __metaclass__ = EdgeMetaClass
    __abstract__ = True

    # if set to True, no more than one edge will
    # be created between two vertices
    __exclusive__ = False

    label = None

    gremlin_path = 'edge.groovy'

    _save_edge = GremlinMethod()
    _get_edges_between = GremlinMethod(classmethod=True)


    FACTORY_CLASS = None
    ## edge id
    ##edge_id = columns.UUID(save_strategy=columns.SAVE_ONCE)

    def __init__(self, outV, inV, **values):
        """
        Initialize this edge with the outgoing and incoming vertices as well as edge properties.

        @param outV: The vertex this edge is coming out of
            @type outV: Vertex
        @param inV: The vertex this edge is going into
            @type inV: Vertex
        @param values: The properties for this edge
            @type values: dict

        """
        self._outV = outV
        self._inV = inV
        super(Edge, self).__init__(**values)

    @classmethod
    def all(cls, ids, as_dict=False):
        """
        Load all edges with the given edge_ids from the graph. By default this will return a list of edges but if
        as_dict is True then it will return a dictionary containing edge_ids as keys and edges found as values.

        @param ids: A list of titan IDs
            @type ids: list
        @param as_dict: Toggle whether to return a dictionary or list
            @type as_dict: boolean
        @returns: dict or list
        """
        if not isinstance(ids, array_types):
            raise MogwaiQueryError("ids must be of type list or tuple")

        strids = [str(i) for i in ids]
        qs = ['ids.collect{g.e(it)}']

        results = execute_query('\n'.join(qs), {'ids': strids})
        results = filter(None, results)

        if len(results) != len(ids):
            raise MogwaiQueryError("the number of results don't match the number of edge ids requested")

        objects = []
        for r in results:
            try:
                objects += [Element.deserialize(r)]
            except KeyError:  # pragma: no cover
                raise MogwaiQueryError('Edge type "%s" is unknown' % '')

        if as_dict:  # pragma: no cover
            return {e._id: e for e in objects}

        return objects

    @classmethod
    def get_label(cls):
        """
        Returns the label for this edge.

        @returns: str

        """
        return cls._type_name(cls.label)

    @classmethod
    def get_between(cls, outV, inV, page_num=None, per_page=None):
        """
        Return all the edges with a given label between two vertices.

        @param outV: The vertex the edge comes out of.
            @type outV: Vertex
        @param inV: The vertex the edge goes into.
            @type inV: Vertex
        @param page_num: The page number of the results
            @type page_num: int
        @param per_page: The number of results per page
            @type per_page : int
        @returns: list

        """
        return cls._get_edges_between(out_v=outV,
                                      in_v=inV,
                                      label=cls.get_label(),
                                      page_num=page_num,
                                      per_page=per_page)

    def validate(self):
        """
        Perform validation of this edge raising a ValidationError if any problems are encountered.
        """
        if self._id is None:
            if self._inV is None:
                raise ValidationError('in vertex must be set before saving new edges')
            if self._outV is None:
                raise ValidationError('out vertex must be set before saving new edges')
        super(Edge, self).validate()

    def save(self, *args, **kwargs):
        """
        Save this edge to the graph database.
        """
        super(Edge, self).save(*args, **kwargs)
        return self._save_edge(self._outV,
                               self._inV,
                               self.get_label(),
                               self.as_save_params(),
                               exclusive=self.__exclusive__)

    def _reload_values(self):
        """ Re-read the values for this edge from the graph database. """
        reloaded_values = {}
        results = execute_query('g.e(id)', {'id': self._id})
        if results:  # note this won't work if you update a node for titan pre-0.5.x, new id's are created
            #del results['_id']
            del results['_type']
            reloaded_values['_id'] = results['_id']
            for name, value in results.get('_properties', {}).items():
                reloaded_values[name] = value
            if results['_id']:
                setattr(self, '_id', results['_id'])
            return reloaded_values
        else:
            return {}

    @classmethod
    def get(cls, id):
        """
        Look up edge by titan assigned ID. Raises a DoesNotExist exception if a edge with the given edge id was not
        found. Raises a MultipleObjectsReturned exception if the edge_id corresponds to more than one edge in the graph.

        @param id: The titan assigned ID
            @type id: str | basestring
        @returns: mogwai.models.Edge
        """
        try:
            results = cls.all([id])
            if len(results) > 1:  # pragma: no cover
                # This requires titan to be broken.
                raise cls.MultipleObjectsReturned

            result = results[0]
            if not isinstance(result, cls):
                raise cls.WrongElementType(
                    '%s is not an instance or subclass of %s' % (result.__class__.__name__, cls.__name__)
                )
            return result
        except MogwaiQueryError:
            raise cls.DoesNotExist

    @classmethod
    def create(cls, outV, inV, label=None, *args, **kwargs):
        """
        Create a new edge of the current type coming out of vertex outV and going into vertex inV with the given
        properties.

        @param outV: The vertex the edge is coming out of
            @type outV: Vertex
        @param inV: The vertex the edge is going into
            @type inV: Vertex

        """
        return super(Edge, cls).create(outV, inV, *args, **kwargs)

    def delete(self):
        """
        Delete the current edge from the graph.
        """
        if self.__abstract__:  # pragma: no cover
            raise MogwaiQueryError('cant delete abstract elements')
        if self._id is None:
            return self
        query = """
        e = g.e(id)
        if (e != null) {
          g.removeEdge(e)
          g.stopTransaction(SUCCESS)
        }
        """
        results = execute_query(query, {'id': self.id})

    def _simple_traversal(self, operation):
        """
        Perform a simple traversal starting from the current edge returning a list of results.

        @param operation: The operation to be performed
            @type operation: str
        @returns: list

        """
        results = execute_query('g.e(id).%s()' % operation, {'id': self.id})
        return [Element.deserialize(r) for r in results]

    def inV(self):
        """
        Return the vertex that this edge goes into.

        @returns: Vertex

        """
        from vertex import Vertex

        if self._inV is None:
            self._inV = self._simple_traversal('inV')[0]
        if isinstance(self._inV, string_types + integer_types):
            self._inV = Vertex.get(self._inV)
        return self._inV

    def outV(self):
        """
        Return the vertex that this edge is coming out of.

        :rtype: Vertex

        """
        from vertex import Vertex

        if self._outV is None:
            self._outV = self._simple_traversal('outV')[0]
        if isinstance(self._inV, string_types + integer_types):
            self._outV = Vertex.get(self._outV)
        return self._outV


from __future__ import unicode_literals
from nose.plugins.attrib import attr

from mogwai.tests.base import BaseMogwaiTestCase, TestVertexModel, TestEdgeModel
from mogwai.exceptions import ModelException, MogwaiException, ValidationError
from mogwai.models import Vertex, Edge
from mogwai import properties


class WildDBNames(Vertex):
    name = properties.String(db_field="words_and_whatnot")
    test_val = properties.Integer(db_field="integers_etc")


@attr('unit', 'class_construction')
class TestModelClassFunction(BaseMogwaiTestCase):
    """
    Tests verifying the behavior of the Model metaclass
    """

    def test_graph_property_attributes_handled_correctly(self):
        """
        Tests that graph property attributes are moved to a _properties dict and replaced with simple value attributes
        """

        #check class attributes
        self.assertHasAttr(TestVertexModel, '_properties')
        self.assertHasAttr(TestVertexModel, 'name')
        self.assertHasAttr(TestVertexModel, 'test_val')

        #check instance attributes
        inst = TestVertexModel()
        self.assertHasAttr(inst, 'name')
        self.assertHasAttr(inst, 'test_val')
        self.assertIsNone(inst.name)
        self.assertIsNone(inst.test_val)

    def test_db_map(self):
        """
        Tests that the db_map is properly defined

        The db_map allows graph properties to be named something other than their variable name
        """

        db_map = WildDBNames._db_map
        self.assertIn('wilddbnames_words_and_whatnot', db_map)
        self.assertIn('wilddbnames_integers_etc', db_map)
        self.assertEquals(db_map['wilddbnames_words_and_whatnot'], 'name')
        self.assertEquals(db_map['wilddbnames_integers_etc'], 'test_val')

    def test_attempting_to_make_duplicate_column_names_fails(self):
        """
        Tests that trying to create conflicting db column names will fail
        """

        with self.assertRaises(ModelException):
            class BadNames(Vertex):
                words = properties.Text()
                content = properties.Text(db_field='words')

    def test_value_managers_are_keeping_model_instances_isolated(self):
        """
        Tests that instance value managers are isolated from other instances
        """
        inst1 = TestVertexModel(test_val=5)
        inst2 = TestVertexModel(test_val=7)

        self.assertNotEquals(inst1.test_val, inst2.test_val)
        self.assertEquals(inst1.test_val, 5)
        self.assertEquals(inst2.test_val, 7)


@attr('unit', 'class_construction')
class TestManualTableNaming(BaseMogwaiTestCase):

    def test_proper_table_naming(self):
        self.assertEqual(TestVertexModel.element_type, 'test_vertex_model')
        self.assertEqual(TestVertexModel.get_element_type(), 'test_vertex_model')


class BaseAbstractVertex(Vertex):
    __abstract__ = True
    data = properties.Text()


class DerivedAbstractVertex(BaseAbstractVertex):
    pass


@attr('unit', 'class_construction')
class TestAbstractElementAttribute(BaseMogwaiTestCase):

    def test_abstract_property_is_not_inherited(self):
        self.assertTrue(BaseAbstractVertex.__abstract__)
        self.assertFalse(DerivedAbstractVertex.__abstract__)

    def test_abstract_element_persistence_methods_fail(self):
        bm = BaseAbstractVertex(data='something')

        with self.assertRaises(MogwaiException):
            bm.save()

        with self.assertRaises(MogwaiException):
            bm.delete()

        with self.assertRaises(MogwaiException):
            bm.update(data='something else')


from __future__ import unicode_literals
from six import print_
import json

from mogwai.connection import generate_spec, execute_query


def get_existing_indices():
    vertex_indices = execute_query('g.getIndexedKeys(Vertex.class)')
    edge_indices = execute_query('g.getIndexedKeys(Edge.class)')
    return vertex_indices, edge_indices


def write_diff_indices_to_file(filename, spec=None):  # pragma: no cover
    if not spec:
        print_("Generating Specification...")
        spec = generate_spec()
    print_("Writing Compiled Diff Indices to File %s ..." % filename)
    vertex_indices, edge_indices = get_existing_indices()
    with open(filename, 'wb') as f:
        for s in spec:
            for pn, pv in s['properties'].items():
                if s['element_type'] == 'Edge' and pn not in edge_indices:
                    f.writelines([pv['compiled'], ])
                elif s['element_type'] == 'Vertex' and pn not in vertex_indices:
                    f.writeliness([json.dumps(pv['compiled']), ])


def write_compiled_indices_to_file(filename, spec=None):  # pragma: no cover
    if not spec:
        print_("Generating Specification...")
        spec = generate_spec()
    print_("Writing Compiled Indices to File %s ..." % filename)
    with open(filename, 'wb') as f:
        for s in spec:
            for pn, pv in s['properties'].items():
                f.writelines([json.dumps(pv['compiled']), ])


def write_specs_to_file(filename):  # pragma: no cover
    print_("Generating Specification...")
    spec = generate_spec()
    print_("Writing Specification to File %s ..." % filename)
    with open(filename, 'wb') as f:
        json.dump(spec, f)
    write_compiled_indices_to_file(filename+'.idx', spec=spec)
    write_compiled_indices_to_file(filename+'.idxdiff', spec=spec)


from vertex import Vertex
from edge import Edge
from paginated_vertex import PaginatedVertex
from query import Query

# in blueprints this is part of the Query.compare
# see http://www.tinkerpop.com/docs/javadocs/blueprints/2.2.0/
EQUAL = "EQUAL"
GREATER_THAN = "GREATER_THAN"
GREATER_THAN_EQUAL = "GREATER_THAN_EQUAL"
LESS_THAN = "LESS_THAN"
LESS_THAN_EQUAL = "LESS_THAN_EQUAL"
NOT_EQUAL = "NOT_EQUAL"

# direction
OUT = "OUT"
IN = "IN"
BOTH = "BOTH"


from __future__ import unicode_literals
from nose.plugins.attrib import attr
from base_tests import GraphPropertyBaseClassTestCase
from mogwai.properties.properties import Integer, Short, PositiveInteger
from mogwai.models import Vertex
from mogwai._compat import print_


@attr('unit', 'property', 'property_int')
class IntegerPropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = Integer
    good_cases = (1, 0)
    bad_cases = ('', 'a', 1.1, None, [], [1], {}, {'a': 1})


class IntegerTestVertex(Vertex):
    element_type = 'test_integer_vertex'

    test_val = Integer()


@attr('unit', 'property', 'property_int')
class IntegerVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_integer_io(self):
        print_("creating vertex")
        dt = IntegerTestVertex.create(test_val=1)
        print_("getting vertex from vertex: %s" % dt)
        dt2 = IntegerTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = IntegerTestVertex.create(test_val=2)
        print_("\ncreated vertex: %s" % dt)
        dt2 = IntegerTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, 2)
        print_("deleting vertex")
        dt2.delete()


@attr('unit', 'property', 'property_short')
class ShortPropertyTestCase(IntegerPropertyTestCase):
    klass = Short


class ShortTestVertex(Vertex):
    element_type = 'test_short_vertex'

    test_val = Short()


@attr('unit', 'property', 'property_short')
class ShortVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_short_io(self):
        print_("creating vertex")
        dt = ShortTestVertex.create(test_val=1)
        print_("getting vertex from vertex: %s" % dt)
        dt2 = ShortTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = ShortTestVertex.create(test_val=2)
        print_("\ncreated vertex: %s" % dt)
        dt2 = ShortTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, 2)
        print_("deleting vertex")
        dt2.delete()


@attr('unit', 'property', 'property_posint')
class PositiveIntegerPropertyTestCase(IntegerPropertyTestCase):
    klass = PositiveInteger


class PositiveIntegerTestVertex(Vertex):
    element_type = 'test_positive_integer_vertex'

    test_val = PositiveInteger()


@attr('unit', 'property', 'property_posint')
class PositiveIntegerVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_positive_integer_io(self):
        print_("creating vertex")
        dt = PositiveIntegerTestVertex.create(test_val=1)
        print_("getting vertex from vertex: %s" % dt)
        dt2 = PositiveIntegerTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = PositiveIntegerTestVertex.create(test_val=2)
        print_("\ncreated vertex: %s" % dt)
        dt2 = PositiveIntegerTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, 2)
        print_("deleting vertex")
        dt2.delete()


from __future__ import unicode_literals
from nose.plugins.attrib import attr

from mogwai.connection import MogwaiQueryError
from mogwai.tests.base import BaseMogwaiTestCase
from mogwai.models import Query, IN, OUT, Edge, Vertex, GREATER_THAN
from mogwai.properties import Integer, Double


class MockVertex(object):
    eid = 1


class MockVertex2(Vertex):
    age = Integer()


class MockEdge(Edge):
    age = Integer()
    fierceness = Double()


@attr('unit', 'query_vertex')
class SimpleQueryTest(BaseMogwaiTestCase):
    def setUp(self):
        self.q = Query(MockVertex())

    def test_limit(self):
        result = self.q.limit(10)._get_partial()
        self.assertEqual(result, "g.v(id).query().limit(limit)")

    def test_direction_in(self):
        result = self.q.direction(IN)._get_partial()
        self.assertEqual(result, "g.v(id).query().direction(IN)")

    def test_direction_out(self):
        result = self.q.direction(OUT)._get_partial()
        self.assertEqual(result, "g.v(id).query().direction(OUT)")

    def test_labels(self):
        result = self.q.labels('test')._get_partial()
        self.assertEqual(result, "g.v(id).query().labels('test')")
        # ensure the original wasn't modified
        self.assertListEqual(self.q._labels, [])

    def test_2labels(self):
        result = self.q.labels('test', 'test2')._get_partial()
        self.assertEqual(result, "g.v(id).query().labels('test', 'test2')")

    def test_object_label(self):
        result = self.q.labels(MockEdge)._get_partial()
        self.assertEqual(result, "g.v(id).query().labels('mock_edge')")

    def test_has(self):
        result = self.q.has(MockEdge.get_property_by_name("age"), 10)._get_partial()
        self.assertEqual(result, "g.v(id).query().has('mockedge_age', v0, Query.Compare.EQUAL)")

    def test_has_double_casting(self):
        result = self.q.has(MockEdge.get_property_by_name("fierceness"), 3.3)._get_partial()
        self.assertEqual(result, "g.v(id).query().has('mockedge_fierceness', v0 as double, Query.Compare.EQUAL)")

    def test_direction_except(self):
        with self.assertRaises(MogwaiQueryError):
            self.q.direction(OUT).direction(OUT)

    def test_has_double_casting_plain(self):
        result = self.q.has('fierceness', 3.3)._get_partial()
        self.assertEqual(result, "g.v(id).query().has('fierceness', v0 as double, Query.Compare.EQUAL)")

    def test_has_int(self):
        result = self.q.has('age', 21, GREATER_THAN)._get_partial()
        self.assertEqual(result, "g.v(id).query().has('age', v0, Query.Compare.GREATER_THAN)")

    def test_intervals(self):
        result = self.q.interval('age', 10, 20)._get_partial()
        self.assertEqual(result, "g.v(id).query().interval('age', v0, v1)")

    def test_double_interval(self):
        result = self.q.interval('fierceness', 2.5, 5.2)._get_partial()
        self.assertEqual(result, "g.v(id).query().interval('fierceness', v0 as double, v1 as double)")



from __future__ import unicode_literals
from nose.plugins.attrib import attr
from base_tests import GraphPropertyBaseClassTestCase
from decimal import Decimal as _D
from mogwai.properties.properties import Decimal
from mogwai.models import Vertex
from mogwai._compat import print_


@attr('unit', 'property', 'property_decimal')
class DecimalPropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = Decimal
    good_cases = (1.1, 0.0, _D(1.1))
    bad_cases = (0, 'val', ['val'], {'val': 1}, None)


class DecimalTestVertex(Vertex):
    element_type = 'test_decimal_vertex'

    test_val = Decimal()


@attr('unit', 'property', 'property_decimal')
class DecimalVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_decimal_io(self):
        print_("creating vertex")
        dt = DecimalTestVertex.create(test_val=_D('1.00'))
        print_("getting vertex from vertex: %s" % dt)
        dt2 = DecimalTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = DecimalTestVertex.create(test_val=5)
        print_("\ncreated vertex: %s" % dt)
        dt2 = DecimalTestVertex.get(dt._id)
        print_("Got decimal vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, _D('5'))
        print_("deleting vertex")
        dt2.delete()



import sys
from setuptools import setup, find_packages

#next time:
#python setup.py register
#python setup.py sdist upload

version = open('mogwai/VERSION', 'r').readline().strip()

long_desc = """
mogwai is an Object-Graph Mapper (OGM) for Python
"""

setup(
    name='mogwai',
    version=version,
    description='Titan Object-Graph Mapper (OGM)',
    dependency_links=['https://bitbucket.org/wellaware/mogwai/archive/{0}.tar.gz#egg=mogwai-{0}'.format(version)],
    long_description=long_desc,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Environment :: Other Environment",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: Implementation",
        "Topic :: Database",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords='cassandra,titan,ogm,mogwai,thunderdome',
    install_requires=['pyparsing>=1.5.7', 'rexpro>=0.1.1', 'six>=1.5.2', 'argparse>=1.2.1', 'factory-boy>=2.3.1'],
    extras_require={
        'metrics': ['pyformance==0.2.4'],
        'develop': ['nose==1.3.0', 'coverage==3.7.1', 'pyformance==0.2.4', 'tox==1.7.1'],
        'newrelic': ['newrelic==2.18.1.15']
    },
    test_suite='nose.collector',
    tests_require=['nose==1.3.0', 'coverage==3.7.1', 'pyformance==0.2.4', 'tox==1.7.1'],
    #setup_requires=['nose==1.3.0', 'coverage==3.7.1', 'pyformance==0.2.4', 'tox==1.7.1'],
    #scripts=['run_converage.sh', 'run_tests.sh'],
    author='Cody Lee',
    author_email='codylee@wellaware.us',
    maintainer='Cody Lee',
    maintainer_email='codylee@wellaware.us',
    url='https://bitbucket.org/wellaware/mogwai',
    license='Apache Software License 2.0',
    packages=find_packages(),
    include_package_data=True,
)



from __future__ import unicode_literals
from nose.plugins.attrib import attr
from mogwai.metrics.manager import MetricManager
from mogwai.tests.base import BaseMogwaiTestCase
from mogwai.metrics.base import BaseMetricsReporter, MetricsRegistry
from mogwai.exceptions import MogwaiMetricsException


@attr('unit', 'metrics')
class MetricManagerTestCase(BaseMogwaiTestCase):
    """
    Test Metric Manager
    """

    def test_default_construction(self):
        mm = MetricManager()
        self.assertEqual(mm.metric_reporters, [])
        with self.assertRaises(MogwaiMetricsException):
            mm.setup_reporters(None)
        mr = BaseMetricsReporter()
        mm.setup_reporters(mr)

        self.assertIsInstance(mm.metric_reporters, (tuple, list))
        self.assertEqual(len(mm.metric_reporters), 1)
        self.assertIsInstance(mm.metric_reporters[0], BaseMetricsReporter)
        self.assertEqual(mm.metric_reporters, [mr])

    def test_multiple_reporter_constuction(self):
        mm = MetricManager()
        r1 = [MetricsRegistry(), MetricsRegistry()]
        mr1 = BaseMetricsReporter(registry=r1)
        r2 = [MetricsRegistry(), MetricsRegistry()]
        mr2 = BaseMetricsReporter(registry=r2)
        mm.setup_reporters([mr1, mr2])

        self.assertIsInstance(mm.metric_reporters, (tuple, list))
        self.assertEqual(len(mm.metric_reporters), 2)
        self.assertIsInstance(mm.metric_reporters[0], BaseMetricsReporter)
        self.assertIsInstance(mm.metric_reporters[1], BaseMetricsReporter)
        self.assertEqual(mm.metric_reporters, [mr1, mr2])

    def test_bad_construction(self):
        mm = MetricManager()

        class NotMetricReporter(object):
            pass

        mr = [NotMetricReporter()]
        with self.assertRaises(MogwaiMetricsException):
            mm.setup_reporters(mr)

    def test_start_stop_mechanism(self):
        mm = MetricManager()
        mr = BaseMetricsReporter()
        mm.setup_reporters(mr)
        mm.start()
        for mr in mm.metric_reporters:
            self.assertTrue(mr.task.running)
        mm.stop()
        for mr in mm.metric_reporters:
            self.assertFalse(mr.task.running)

    def test_metric_getter(self):
        mm = MetricManager()
        mr = BaseMetricsReporter()
        mm.setup_reporters(mr)
        mm.start()
        # increment counters
        for counter in mm.counters('test'):
            counter.inc()

        timestamp, metrics = mr.get_metrics()
        self.assertIsInstance(timestamp, (int, long))
        self.assertIsInstance(metrics, dict)
        self.assertEqual(len(metrics), 1)
        self.assertIn('test', metrics)
        self.assertIsInstance(metrics['test'], dict)
        self.assertEqual(metrics['test']['count'], 1)
        mm.stop()

    def test_counter_decorator(self):
        mm = MetricManager()
        mr = BaseMetricsReporter()
        mm.setup_reporters(mr)
        mm.start()

        @mm.count_calls
        def somefunc(i):
            return i

        @mm.count_calls
        def badfunc(i):
            raise Exception("test exception")

        self.assertEqual(somefunc(1), 1)
        self.assertEqual(somefunc(1, context='test'), 1)
        with self.assertRaises(Exception):
            badfunc(1)

        # check metrics collected
        for reporter in mm.metric_reporters:
            timestamp, metrics = reporter.get_metrics()
            self.assertIsInstance(timestamp, (int, long))
            self.assertIsInstance(metrics, dict)
            self.assertEqual(len(metrics), 5)

            # context aware
            self.assertIn('test.counter', metrics)
            self.assertIsInstance(metrics['test.counter'], dict)
            self.assertEqual(metrics['test.counter']['count'], 1)

            # normal one
            self.assertIn('somefunc.calls', metrics)
            self.assertIsInstance(metrics['somefunc.calls'], dict)
            self.assertEqual(metrics['somefunc.calls']['count'], 2)

        mm.stop()

    def test_timer_decorator(self):
        mm = MetricManager()
        mr = BaseMetricsReporter()
        mm.setup_reporters(mr)
        mm.start()

        @mm.time_calls
        def somefunc(i):
            return i

        @mm.time_calls
        def badfunc(i):
            raise Exception("test exception")

        self.assertEqual(somefunc(1), 1)
        self.assertEqual(somefunc(1, context='test'), 1)
        with self.assertRaises(Exception):
            badfunc(1)

        # check metrics collected
        for reporter in mm.metric_reporters:
            timestamp, metrics = reporter.get_metrics()

            self.assertIsInstance(timestamp, (int, long))
            self.assertIsInstance(metrics, dict)
            self.assertEqual(len(metrics), 5)

            # context aware
            self.assertIn('test.timer', metrics)
            self.assertIsInstance(metrics['test.timer'], dict)
            self.assertLess(metrics['test.timer']['max'], 0.01)

            # normal one
            self.assertIn('somefunc.timer', metrics)
            self.assertIsInstance(metrics['somefunc.timer'], dict)
            self.assertLess(metrics['somefunc.timer']['max'], 0.01)

        mm.stop()

    def test_meter_decorator(self):
        mm = MetricManager()
        mr = BaseMetricsReporter()
        mm.setup_reporters(mr)
        mm.start()

        @mm.meter_calls
        def somefunc(i):
            return i

        @mm.meter_calls
        def badfunc(i):
            raise Exception("test exception")

        self.assertEqual(somefunc(1), 1)
        self.assertEqual(somefunc(1, context='test'), 1)
        with self.assertRaises(Exception):
            badfunc(1)

        # check metrics collected
        for reporter in mm.metric_reporters:
            timestamp, metrics = reporter.get_metrics()
            self.assertIsInstance(timestamp, (int, long))
            self.assertIsInstance(metrics, dict)
            self.assertEqual(len(metrics), 5)

            # context aware
            self.assertIn('test.meter', metrics)
            self.assertIsInstance(metrics['test.meter'], dict)
            self.assertGreater(metrics['test.meter']['mean_rate'], 100)

            # normal one
            self.assertIn('somefunc.calls', metrics)
            self.assertIsInstance(metrics['somefunc.calls'], dict)
            self.assertGreater(metrics['somefunc.calls']['mean_rate'], 100)

        mm.stop()

    def test_histogram_decorator(self):
        mm = MetricManager()
        mr = BaseMetricsReporter()
        mm.setup_reporters(mr)
        mm.start()

        @mm.hist_calls
        def somefunc(i):
            return i

        @mm.hist_calls
        def badfunc(i):
            raise Exception("test exception")

        self.assertEqual(somefunc(1), 1)
        self.assertEqual(somefunc(1, context='test'), 1)
        with self.assertRaises(Exception):
            badfunc(1)

        # check metrics collected
        for reporter in mm.metric_reporters:
            timestamp, metrics = reporter.get_metrics()
            self.assertIsInstance(timestamp, (int, long))
            self.assertIsInstance(metrics, dict)
            self.assertEqual(len(metrics), 4)

            # context aware
            self.assertIn('test.hist', metrics)
            self.assertIsInstance(metrics['test.hist'], dict)
            self.assertEqual(metrics['test.hist']['count'], 1)

            # normal one
            self.assertIn('somefunc.calls', metrics)
            self.assertIsInstance(metrics['somefunc.calls'], dict)
            self.assertEqual(metrics['somefunc.calls']['count'], 2)

        mm.stop()

    def test_multiple_decorators(self):
        mm = MetricManager()
        mr = BaseMetricsReporter()
        mm.setup_reporters(mr)
        mm.start()

        @mm.meter_calls
        @mm.count_calls
        @mm.hist_calls
        @mm.time_calls
        def somefunc(i):
            return i

        self.assertEqual(somefunc(1), 1)
        self.assertEqual(somefunc(1, context='test'), 1)

        # check metrics collected
        for reporter in mm.metric_reporters:
            timestamp, metrics = reporter.get_metrics()
            self.assertIsInstance(timestamp, (int, long))
            self.assertIsInstance(metrics, dict)
            self.assertEqual(len(metrics), 3)

            # context aware
            self.assertIn('test.meter', metrics)
            self.assertIsInstance(metrics['test.meter'], dict)
            self.assertGreater(metrics['test.meter']['mean_rate'], 100)

            # normal one
            self.assertIn('somefunc.calls', metrics)
            self.assertIsInstance(metrics['somefunc.calls'], dict)
            self.assertEqual(metrics['somefunc.calls']['count'], 2)

        mm.stop()

    def test_get_all_timers(self):
        mm = MetricManager()
        mr = BaseMetricsReporter()
        mm.setup_reporters(mr)

        # test default implementation
        for timer in mm.timers():
            self.assertIsNotNone(timer.im_self)
            self.assertIsInstance(timer.im_self, MetricsRegistry)

        # test context aware
        for timer in mm.timers('test'):
            from pyformance.meters.timer import Timer
            self.assertIsInstance(timer, Timer)

    def test_get_all_histograms(self):
        mm = MetricManager()
        mr = BaseMetricsReporter()
        mm.setup_reporters(mr)

        # test default implementation
        for hist in mm.histograms():
            self.assertIsNotNone(hist.im_self)
            self.assertIsInstance(hist.im_self, MetricsRegistry)

        # test context aware
        for hist in mm.histograms('test'):
            from pyformance.meters.histogram import Histogram
            self.assertIsInstance(hist, Histogram)

    def test_get_all_meters(self):
        mm = MetricManager()
        mr = BaseMetricsReporter()
        mm.setup_reporters(mr)

        # test default implementation
        for meter in mm.meters():
            self.assertIsNotNone(meter.im_self)
            self.assertIsInstance(meter.im_self, MetricsRegistry)

        # test context aware
        for meter in mm.meters('test'):
            from pyformance.meters.meter import Meter
            self.assertIsInstance(meter, Meter)

    def test_get_all_counters(self):
        mm = MetricManager()
        mr = BaseMetricsReporter()
        mm.setup_reporters(mr)

        # test default implementation
        for counter in mm.counters():
            self.assertIsNotNone(counter.im_self)
            self.assertIsInstance(counter.im_self, MetricsRegistry)

        # test context aware
        for counter in mm.counters('test'):
            from pyformance.meters.counter import Counter
            self.assertIsInstance(counter, Counter)


from __future__ import unicode_literals
from nose.plugins.attrib import attr
from mogwai.tests import BaseMogwaiTestCase
from base_tests import GraphPropertyBaseClassTestCase
from mogwai.properties.properties import Boolean, GraphProperty
from mogwai.models import Vertex
from mogwai.exceptions import ValidationError
from mogwai._compat import print_


@attr('unit', 'property', 'property_boolean')
class BooleanPropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = Boolean
    good_cases = (True, False)
    bad_cases = (0, 1.1, 'val', [], (), {}, None)


class BooleanTestVertex(Vertex):
    element_type = 'test_boolean_vertex'

    test_val = Boolean()

CHOICES = (
    (True, True),
    (False, False)
)


class BooleanTestChoicesVertex(Vertex):
    element_type = 'test_boolean_choices_vertex'

    test_val = Boolean(choices=CHOICES)


@attr('unit', 'property', 'property_boolean')
class BooleanVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_boolean_io(self):
        print_("creating vertex")
        dt = BooleanTestVertex.create(test_val=True)
        print_("getting vertex from vertex: %s" % dt)
        dt2 = BooleanTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = BooleanTestVertex.create(test_val=True)
        print_("\ncreated vertex: %s" % dt)
        dt2 = BooleanTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, True)
        print_("deleting vertex")
        dt2.delete()


@attr('unit', 'property', 'property_boolean')
class TestVertexChoicesTestCase(BaseMogwaiTestCase):

    def test_good_choices_value_io(self):
        print_("creating vertex")
        dt = BooleanTestChoicesVertex.create(test_val=True)
        print_("validating input")
        self.assertEqual(dt.test_val, True)
        print_("deleting vertex")
        dt.delete()

    def test_good_choices_key_io(self):
        print_("creating vertex")
        dt = BooleanTestChoicesVertex.create(test_val=False)
        print_("validating input")
        self.assertEqual(dt.test_val, False)
        print_("deleting vertex")
        dt.delete()

    def test_bad_choices_io(self):
        with self.assertRaises(ValidationError):
            print_("creating vertex")
            dt = BooleanTestChoicesVertex.create(test_val=None)
            print_("validating input")
            self.assertEqual(dt.test_val, None)
            print_("deleting vertex")
            dt.delete()


from __future__ import unicode_literals
from nose.plugins.attrib import attr
from base_tests import GraphPropertyBaseClassTestCase
from mogwai.properties.properties import Email
from mogwai.models import Vertex
from mogwai._compat import print_


@attr('unit', 'property', 'property_email')
class EmailPropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = Email
    good_cases = ('test@test.com', 'test-alice@subdomain.domain.com')
    bad_cases = ('', 'bob', '@test.com', 'bob@test')

    def test_email_edge_cases(self):
        p = Email(default='bob@bob.net')


class EmailTestVertex(Vertex):
    element_type = 'test_email_vertex'

    test_val = Email()


@attr('unit', 'property', 'property_email')
class EmailVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_email_io(self):
        print_("creating vertex")
        dt = EmailTestVertex.create(test_val='alice@alice.com')
        print_("getting vertex from vertex: %s" % dt)
        dt2 = EmailTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = EmailTestVertex.create(test_val='bob@bob.com')
        print_("\ncreated vertex: %s" % dt)
        dt2 = EmailTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, 'bob@bob.com')
        print_("deleting vertex")
        dt2.delete()